#! /usr/bin/python -O
# -*- python -*-

# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# launch this script to generate the level

import math, os, os.path
import copy

import Tkinter

import soya
import soya.model  as model
import soya.soya3d as soya3d
import soya.editor.main
import soya.land

import slune.level

import circuit_generator
import town_generator

soya.init()



def make_land():

    land = soya.land.Land()
    land.from_image(model.Image("/home/blam/data/bombay2.png"))
    land.map_size = 8
    land.scale_factor = 1.5
    land.multiply_height(35.0)
    land.add_height(-10.0)
    land.set_texture_layer(model.Material.get("ground4"), 0.0,  150.0)
    land.set_texture_layer(model.Material.get("rocks3"), 7.0, 150.0)
    land.set_texture_layer_angle(model.Material.get("sand1"), 0.0, 150.0, 0.0, 10.0)
    land.set_texture_layer_angle(model.Material.get("ground5"), 0.0, 4.0, 0.0, 10.0)
    land.set_texture_layer_angle(model.Material.get("rocks3"), 0.0, 150.0, 50.0, 90.0)

    # make wholes for metro

    def make_whole (x, y, w, h):
        j = 0
        while j < h:
            i = 0
            while i < w:
                land.add_point_option(x + i, y + j, 1 | 4 | 8)
                i += 1
            j += 1

    def add_height (x, y, h):
        land.set_height(x, y, land.get_height (x, y) + h)

    def add_height_area (x, y, w, h, a):
        j = 0
        while j < h:
            i = 0
            while i < w:
                land.set_height(x + i, y + j, land.get_height (x + i, y + j) + a)
                i += 1
            j += 1

    # whole 1
    add_height_area (50, 85, 4, 2, 0.4)
    add_height_area (50, 86, 4, 1, 0.2)
    add_height(50, 86, -0.2)
    make_whole(50, 86, 4, 3)

    add_height(54, 88, -0.4)
    add_height(53, 88, -0.6)
    add_height(52, 88, -0.4)

    # whole 2
    land.set_height(66, 44, land.get_height (66, 44) + 0.1)
    land.set_height(67, 45, land.get_height (67, 45) + 0.4)
    land.set_height(68, 45, land.get_height (68, 45) + 0.5)
    land.set_height(67, 44, land.get_height (67, 44) + 0.4)
    land.set_height(68, 44, land.get_height (68, 44) + 0.4)
    add_height(66, 44, -0.1)
    add_height(65, 43, -1.0)
    land.add_point_option(65, 43, 1 | 4 | 8)
    land.add_point_option(65, 44, 1 | 4 | 8)
    land.add_point_option(67, 45, 1 | 4 | 8)
    land.add_point_option(66, 44, 1 | 4 | 8)
    land.add_point_option(66, 45, 1 | 4 | 8)
    land.add_point_option(64, 44, 1 | 4 | 8)
    land.add_point_option(65, 45, 1 | 4 | 8)
    land.add_point_option(66, 46, 1 | 4 | 8)
    land.add_point_option(64, 45, 1 | 4 | 8)
    land.add_point_option(65, 46, 1 | 4 | 8)
    land.add_point_option(64, 46, 1 | 4 | 8)
    add_height(64, 46, -1.2)
    add_height(66, 46, 1.0)

    # whole 3
    make_whole(32, 39, 4, 3)
    add_height(36, 39, -1.3)
    add_height(35, 39, -1.8)
    add_height(34, 39, -1.3)
    add_height(33, 39, -0.8)
    add_height(35, 38, -0.5)
    add_height(32, 42, 0.5)
    add_height(33, 42, 0.5)
    add_height(35, 42, -0.5)
    add_height(33, 41, 0.3)
    add_height(35, 41, -0.8)
    add_height(35, 40, -0.4)
    add_height(34, 40, -0.4)

    add_height(31, 38, 0.6)
    add_height(32, 38, 1.0)
    add_height(33, 38, 0.6)

    add_height(31, 37, 0.5)
    add_height(32, 37, 0.5)
    add_height(33, 37, 0.5)

    # whole 4
    make_whole(31, 63, 3, 5)
    make_whole(30, 64, 5, 3)
    
    # planed terrain for labs
    m = model.Material.get("ground6")
    j = 88
    while j <= 93:
        i = 70
        while i <= 82:
            land.set_height(i, j, 1.5)
            land.set_texture(i, j, m)
            i += 1
        j += 1

    # planed terrain for houses
    add_height(80, 41, 0.4)
    add_height(79, 41, 0.64)
    add_height(80, 42, 0.4)
    add_height(78, 41, 0.64)
    add_height(79, 40, 0.64)
    add_height(78, 40, 0.64)
    add_height(78, 42, 0.64)
    add_height(79, 42, 0.64)

    add_height(89, 42, 0.2)

    add_height_area (25, 24, 3, 8, 0.5)
    add_height_area (25, 11, 3, 6, 0.2)

    add_height_area (82, 99, 4, 4, 0.2)

    # roads
    import Image

    circuit = Image.open("/home/blam/data/bombay2_ways.png")
    m = model.Material.get("ground6")
    i = 0
    while(i < 129):
        j = 0
        while(j < 129):
            if circuit.getpixel((i, j)) == (255, 0, 0):
                land.set_texture(i, j, m)
                add_height(i, j, -0.5)
            j = j + 1
        i = i + 1

    return land

land = make_land()

def make_metro():

    w = soya3d.World()
    w.shapify_args = ('tree', {})

    # metro light
    light = soya3d.Light(w)
    light.directional = 1
    light.diffuse = (-0.1, -0.1, -0.1, 1.0)
    light._matrix = (-1.0, -0.0, 0.0, 0.0, -0.0, 0.0, -1.0, 0.0, -0.0, -1.0, -0.0, 0.0, 0.0, 5.0, 0.0, 1.0, 1.0, 1.0, 1.0)

    # add tubes
    v = soya3d.World(w)
    s = model.Shape.get('metro_all')
    s.to_faces(v)

    w.filename = "shape-cyberbombay-1"
    w.save()

    return w


def make_houses():
    w = soya3d.World()
    houses = soya3d.World(w)

    w.shapify_args = ('tree', {})

    # houses light
    light = soya3d.Light(w)
    light.directional = 1
    light.static = 1
    light._matrix = (0.63844746351242065, 0.0, 0.76966547966003418, 0.0, 0.33340060710906982, -0.90130943059921265, -0.27656009793281555, 0.0, 0.6937066912651062, 0.43317598104476929, -0.57543867826461792, 0.0, 0.0, 5.0, 0.0, 1.0, 1.0, 1.0, 1.0)

    # add entrances
    v = soya3d.World(w)
    s = model.Shape.get('metro_entrance1')
    s.to_faces(v)
    v.set_xyz(77.25, 5.0, 132.6)
    
    v = soya3d.World(w)
    s = model.Shape.get('metro_entrance1')
    s.to_faces(v)
    v.rotate_lateral(135.0)
    v.set_xyz(99.5, 2.8, 65.5)

    v = soya3d.World(w)
    s = model.Shape.get('metro_entrance1')
    s.to_faces(v)
    v.rotate_lateral(180.0)
    v.set_xyz(50.5, 5.2, 57.75)

    v = soya3d.World(w)
    s = model.Shape.get('metro_entrance2')
    s.to_faces(v)
    v.scale(1.5, 1.5, 1.5)
    v.set_xyz(48.0, 11.2, 97.5)

    # houses
    model.Material.PATH = "/home/blam/prog/slune/materials"
    m1 = model.Material.get('house3')
    m2 = model.Material.get('inde_2')
    m3 = model.Material.get('bark1')
    m4 = model.Material.get('house3b')
    m5 = model.Material.get('inde_4')
    m6 = model.Material.get('inde_1')
    m7 = model.Material.get('inde_3')
    m8 = model.Material.get('thatch1')
    m9 = model.Material.get('taj3')
    m10 = model.Material.get('inde_5')
    m11 = model.Material.get('bloc3')
    m12 = model.Material.get('bloc3win')
    m13 = model.Material.get('ground5')
    m14 = model.Material.get('taj1')

    ms1 = (m1, 'plak round', 0.5, 'complete', 1.0)
    ms2 = (m2, 'plak', 1.0, 'plak', 1.0)
    ms3 = (m3, 'plak', 1.0, 'plak', 1.0)
    ms4 = (m4, 'plak round', 1.0, 'complete', 1.0)
    ms6 = (m6, 'plak round', 1.0, 'complete', 1.0)
    ms7 = (m7, 'plak round', 1.0, 'complete', 1.0)
    ms8 = (m9, 'plak', 1.0, 'complete', 2.0)
    ms9 = (m9, 'plak', 1.0, 'complete', 1.0)
    ms10 = (m10, 'plak round', 1.0, 'complete', 1.0)
    ms11 = (m11, 'plak round', 1.0, 'complete', 1.0)
    ms12 = (m12, 'plak round', 1.0, 'complete', 1.0)
    ms13 = (m4, 'plak round', 1.0, 'complete', 2.0)
    ms14 = (m14, 'plak round', 1.0, 'complete', 2.0)
    ms15 = (m6, 'plak round', 1.0, 'complete', 2.0)

    def make_house(points, style, x = 0.0, z = 0.0, angle = 0.0, add_y = 0.0):
        v = soya3d.World(houses)
        h = town_generator.House(points)
        if style == 1:
            h.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 3.0), (0.2, 3.2), (0.2, 4.2), (0.4, 4.3)]
            h.materials = [ ms2, ms1, ms3, ms4, ms3 ]
            h.roof = ('trapez', m5, m5, 1.5, 2.0, 0.5, 1.0)
        elif style == 2:
            h.section = [(0.2, 0.0), (0.2, 1.0), (0.2, 2.0), (0.0, 2.2), (0.0, 3.2), (0.0, 4.2)]
            h.materials = [ ms2, ms6, ms3, ms7, ms2 ]
            h.roof = ('trapez', m3, m2, 1.5, 0.0, 1.0, 1.5)
        elif style == 3:
            h.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 3.0), (0.3, 3.3), (0.3, 4.3)]
            h.materials = [ ms6, ms2, ms3, ms7 ]
            h.roof = ('pyramid', m5, 2.0, 0.5, 'complete', 3.0)
        elif style == 4:
            h.section = [(0.3, 0.0), (0.3, 1.0), (0.0, 1.2), (0.0, 2.2), (0.0, 3.2), (0.0, 4.2), (0.0, 5.2), (0.3, 5.4), (0.3, 6.4)]
            h.materials = [ ms10, ms2, ms11, ms12, ms11, ms12, ms2, ms10 ]
            h.roof = ('flat', m13)
        elif style == 5:
            h.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 2.0), (0.0, 3.0), (0.3, 3.2), (0.3, 4.2)]
            h.materials = [ ms11, ms12, ms11, ms2, ms10 ]
            h.roof = ('flat', m13)
        elif style == 6:
            h.section = [(0.0, 0.0), (0.0, 1.0), (0.0, 3.0), (0.0, 4.0), (0.0, 6.0), (0.3, 6.2)]
            h.materials = [ ms6, ms13, ms6, ms13, ms3 ]
            h.roof = ('pyramid', m3, 2.5, 0.5, 'complete', 2.0)
        elif style == 7:
            h.section = [(0.0, 0.0), (0.0, 4.0), (0.2, 5.0)]
            h.materials = [ ms14, ms9 ]
            h.roof = ('pyramid', m13, 1.5, 0.5, 'complete', 2.0)
        elif style == 8:
            h.section = [(0.0, 0.0), (0.0, 2.0), (-0.2, 2.2), (-0.2, 4.2), (-0.4, 4.4), (-0.4, 6.4), (-0.6, 6.6), (-0.6, 8.6)]
            h.materials = [ ms15, ms3, ms15, ms3, ms15, ms3, ms15 ]
            h.roof = ('trapez', m5, m2, 1.0, 0.0, 1.0, 2.0)
        h.rotate_points(angle)
        h.translate_points(x, z)
        h.draw(v)
        h.set_on_land(land, v)
        v.y += add_y
        return v

    # add labo
    v = soya3d.World(w)
    s = model.Shape.get('labs1')
    s.to_faces(v)
    v.set_xyz(108.0, 4.24, 135.0)

    ps = [(116.09649658, 132.345474243), (122.88722229, 132.361022949), (122.877716064, 138.14730835), (116.095108032, 138.137893677)]
    h = town_generator.House(ps)
    h.set_on_land(land)
    ps = [(107.111419678, 132.110473633), (112.877716064, 132.110473633), (112.877716064, 134.912887573), (107.104621887, 134.877944946)]
    h = town_generator.House(ps)
    h.set_on_land(land)

    # add temple
    v = soya3d.World(w)
    s = model.Shape.get('pagode_3')
    s.to_faces(v)
    v.rotate_lateral(-90.0)
    v.set_xyz(65.0, 11.2, 98.0)

    ps = [(67.2001876831, 93.2055587769), (75.7270202637, 93.2624893188), (75.7912521362, 105.295455933), (67.1547546387, 105.295455933)]
    h = town_generator.House(ps)
    h.set_on_land(land)

    v = soya3d.World(w)
    s = model.Shape.get('pagode_1')
    s.to_faces(v)
    v.set_xyz(64.0, 12.4, 86.0)

    # rempart
    v = soya3d.World(w)
    s = model.Shape.get('rempart_indus1')
    s.to_faces(v)
    v.rotate_lateral(180.0)
    v.set_xyz(140.0, 1.5, 80.0)
    
    v = soya3d.World(w)
    s = model.Shape.get('rempart_indus4')
    s.to_faces(v)
    v.rotate_lateral(-80.0)
    v.set_xyz(72.0, 0.6, 147.0)

    v = soya3d.World(w)
    s = model.Shape.get('rempart_indus5')
    s.to_faces(v)
    v.set_xyz(50.0, 1.34, 13.0)

    # special house
    v = soya3d.World(w)
    s = model.Shape.get('house_indus1')
    s.to_faces(v)
    v.rotate_lateral(-20.0)
    v.set_xyz(137.0, 1.6, 54.0)
    
    ps = [(132.055343628, 55.507106781), (139.790344238, 58.2436523438), (141.824035645, 52.4744377136), (134.17678833, 49.7211456299)]
    h = town_generator.House(ps)
    h.set_on_land(land)

    v = soya3d.World(w)
    s = model.Shape.get('house_indus2')
    s.to_faces(v)
    v.rotate_lateral(-45.0)
    v.set_xyz(134.0, 3.0, 108.0)

    # generic houses
    make_house(town_generator.make_square(6.0, 10.0), 1, 148.5, 39.0)
    make_house(town_generator.make_square(12.0, 6.0), 2, 136.0, 35.5, 22.0, add_y = -0.2)
    make_house(town_generator.make_square(4.0, 4.0), 4, 126.0, 50.0, 18.0, add_y = -0.2)
    make_house(town_generator.make_square(4.0, 4.0), 4, 119.0, 64.0, 17.0, add_y = -0.1)
    make_house(town_generator.make_square(6.0, 12.0), 1, 109.0, 40.0, add_y = -0.2)
    make_house(town_generator.make_square(12.0, 6.0), 5, 116.0, 20.0)
    make_house(town_generator.make_square(4.0, 4.0), 6, 83.0, 44.5)
    make_house(town_generator.make_square(6.0, 10.0), 6, 133.0, 67.2, add_y = -0.15)
    make_house([(-2.0, 5.0), (2.0, 5.0), (5.0, 2.0), (5.0, -2.0), (2.0, -5.0), (-2.0, -5.0), (-5.0, -2.0), (-5.0, 2.0)], 3, 98.0, 158.0, add_y = -0.1)
    make_house(town_generator.make_square(6.0, 6.0), 3, 128.0, 85.0, add_y = -0.1)
    make_house(town_generator.make_square(6.0, 12.0), 2, 168.0, 70.0, add_y = -0.0)
    make_house(town_generator.make_square(10.0, 8.0), 2, 87.0, 33.0, add_y = -0.0)
    make_house(town_generator.make_square(20.0, 4.0), 7, 82.0, 18.0, add_y = -0.0)
    make_house(town_generator.make_square(4.0, 8.0), 1, 57.0, 37.0, -30.0, add_y = -0.0)
    make_house(town_generator.make_square(4.0, 4.0), 8, 22.0, 50.0, 45.0, add_y = -0.32)
    make_house(town_generator.make_square(14.0, 6.0), 1, 80.0, 174.0, -10.0, add_y = -0.0)
    make_house([(-4.0, 3.0), (4.0, 3.0), (6.0, 0.0), (4.0, -3.0), (-4.0, -3.0), (-6.0, 0.0)], 7, 138.0, 21.0, -40.0, add_y = -0.0)
    make_house(town_generator.make_square(12.0, 10.0), 4, 130.0, 154.0, 0.0, add_y = 0.25)
    make_house(town_generator.make_square(4.0, 4.0), 8, 95.0, 143.0, 45.0, add_y = -0.32)
    make_house(town_generator.make_square(6.0, 14.0), 6, 149.0, 105.0, add_y = -0.0)
    make_house(town_generator.make_square(12.0, 6.0), 2, 141.0, 122.0, add_y = -0.0)
    make_house(town_generator.make_square(8.0, 10.0), 7, 156.0, 128.0, add_y = -0.33)
    make_house(town_generator.make_square(6.0, 12.0), 1, 135.0, 137.0, add_y = -0.0)
    make_house(town_generator.make_square(10.0, 5.0), 2, 150.0, 142.0, add_y = -0.0)
    make_house(town_generator.make_square(4.0, 4.0), 8, 18.0, 122.0, add_y = -0.0)


    # palmtrees
    v = soya3d.World(w)
    s = model.Shape.get('palmtree1')
    s.to_faces(v)
    v.scale(0.4, 0.4, 0.4)
    v.set_xyz(40.0, 0.8, 146.0)

    v = soya3d.World(w)
    s.to_faces(v)
    v.scale(0.3, 0.3, 0.3)
    v.rotate_vertical(30.0)
    v.rotate_lateral(-45.0)
    v.set_xyz(39.5, 0.7, 146.5)

    v = soya3d.World(w)
    s.to_faces(v)
    v.scale(0.4, 0.4, 0.4)
    v.rotate_lateral(90.0)
    v.set_xyz(34.0, 12.3, 85.0)

    v = soya3d.World(w)
    s.to_faces(v)
    v.scale(0.4, 0.4, 0.4)
    v.rotate_lateral(100.0)
    v.set_xyz(167.0, 1.65, 39.0)

    v = soya3d.World(w)
    s.to_faces(v)
    v.scale(0.4, 0.4, 0.4)
    v.rotate_lateral(180.0)
    v.set_xyz(171.0, 0.85, 161.0)

    v = soya3d.World(w)
    s.to_faces(v)
    v.scale(0.4, 0.4, 0.4)
    v.rotate_lateral(180.0)
    v.rotate_incline(-20.0)
    v.set_xyz(162.0, 2.25, 166.0)

    def del_vertices_color(obj):
        if isinstance(obj, soya3d.World):
            for o in obj.children:
                del_vertices_color(o)
        elif isinstance(obj, model.Face):
            for v in obj.vertices:
                v.color = None

    del_vertices_color(w)

# hack
#    w.filename = "shape-cyberbombay-2"
#    w.save()
    return w


def make_world():

    w = soya3d.World()
    w.shape = land
    make_houses()

    atm = soya3d.Atmosphere()
    atm.ambient = (0.45, 0.43, 0.4, 1.0)
    atm.bg_color  = (0.6, 0.6, 0.8, 1.0)
    atm.skyplane = 1
    atm.sky_color = (0.8, 0.8, 1.0, 1.0)
    atm.fog_color = atm.bg_color
    atm.fog_type    = 0
    atm.fog_start   = 20.0
    atm.fog_end     = 50.0
    atm.fog         = 1
    atm.fog_density = 0.001
    atm.cloud = model.Material.get("clouds1")
    w.atmosphere = atm

    # houses light
    light = soya3d.Light(w)
    light.directional = 1
    light.static = 1
    light._matrix = (0.63844746351242065, 0.0, 0.76966547966003418, 0.0, 0.33340060710906982, -0.90130943059921265, -0.27656009793281555, 0.0, 0.6937066912651062, 0.43317598104476929, -0.57543867826461792, 0.0, 0.0, 5.0, 0.0, 1.0, 1.0, 1.0, 1.0)

    # add labo smoke
    import soya
    smoke = soya.Smoke(w)
    smoke.nb_max = 20
    smoke.auto_generate_particle = 1
    smoke.set_xyz(121.9, 4.5, 131.5)
    
    v = soya3d.Volume(w)
    v.shape = model.Shape.get('shape-cyberbombay-2')

    land.compute_static_light([light], 1)

    v = soya3d.World(w)
    v.shape = model.Shape.get('shape-cyberbombay-1')
    atm = soya3d.Atmosphere()
    atm.ambient = (0.3, 0.3, 0.3, 1.0)
    atm.bg_color  = (0.0, 0.0, 0.0, 1.0)
    atm.skyplane = 0
    atm.fog_color = atm.bg_color
    atm.fog_type    = 0
    atm.fog_start   = 20.0
    atm.fog_end     = 50.0
    atm.fog         = 1
    atm.fog_density = 0.001
    v.atmosphere = atm

    w.filename = "world-cyberbombay"
    w.save()
    return w



#print 'shape-cyberbombay-1 ...'
#make_metro()
#print '  [ OK ]'
#print 'shape-cyberbombay-2 ...'
#make_houses()
#print '  [ OK ]'
print 'world-cyberbombay ...'
make_world()
print '  [ OK ]'


#soya.editor.edit(w)
#Tkinter.mainloop()





# Slune
# Copyright (C) 2004 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import random

import soya

import slune.character as slune_character, slune.sound as sound

class AirplaneParticles(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 30):
    soya.Particles.__init__(self, parent, material, nb_particles)
    self.auto_generate_particle = 1
    self.set_colors((1.0, 1.0, 1.0, 1.0), (1.0, 1.0, 0.0, 1.0), (1.0, 0.0, 0.0, 1.0), (0.1, 0.0, 0.0, 1.0))
    self.set_sizes((1.0, 1.0))
    self.max_particles_per_round = 1
    
  def generate(self, index):
    self.set_particle(index, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
  

class Airplane(soya.World):
  def __init__(self, parent):
    soya.World.__init__(self, parent)
    self.shape          = soya.Shape.get("avion")
    self.characters_on  = ()
    self.speed          = soya.Vector(self)
    self.accelerate     = 0
    self.roof           = soya.Point(self, 0.0, 7.9, 0.0)
    self.start_in_round = 0
    self.particles      = ()
    
  def start(self):
    self.accelerate = 1
    self.create_particles()
    
  def create_particles(self):
    if not self.particles:
      self.particles = [AirplaneParticles(self), AirplaneParticles(self)]
      self.particles[0].set_xyz( 1.0, 3.5, 9.6)
      self.particles[1].set_xyz(-1.0, 3.5, 9.6)
      
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    self.add_mul_vector(proportion, self.speed)
    for character in self.characters_on: character.add_mul_vector(proportion, self.speed)
      
  def begin_round(self):
    soya.World.begin_round(self)
    
    if self.accelerate:
      if self.speed.z > -2.0:
        self.speed.z -= 0.003
        if self.speed.z < -1.0: self.turn_vertical(0.05)
        
    self.characters_on = [
      character
      for character in self.parent.characters
      if character.on_ground and (abs(character.y - self.y - self.roof.y) < 2.0) and (character.distance_to(self.roof) < 11.0)
      ]

    for character in self.characters_on: character.platform_speed = self.speed

    if self.speed.z == 0.0:
      if   self.characters_on: self.start()
      elif self.start_in_round:
        self.start_in_round -= 1
        if self.start_in_round == 0: self.start()


class FlyingAirplane(Airplane):
  def __init__(self, parent):
    Airplane.__init__(self, parent)
    self.speed.z = -0.5
    self.create_particles()
    
    
class KillerAirplane(soya.World):
  def __init__(self, parent, target, speed = 0.4):
    soya.World.__init__(self, parent)
    self.volume = soya.Volume(self, soya.Shape.get("avion"))
    self.volume.scale(0.5, 0.5, 0.5)
    self.target = target
    self.phase  = 1
    self.v      = soya.Vector()
    self.p      = soya.Point(self.target, 0.0, 0.0, -4.0)
    self.speed  = speed
    self.solid  = 1
    self.play   = 0
    
  def begin_round(self):
    soya.World.begin_round(self)

    if self.play == 0: return
    
    if   self.phase == 1:
      self.p.convert_to(self.parent)
      
      if (abs(self.x - self.p.x) + abs(self.z - self.p.z)) < 2.0:
        r = self.parent.raypick(self, soya.Vector(self, 0.0, -1.0, 0.0))
        if r: 
          self.drop_y = max(r[0].y, 0.0)
          print self.drop_y
          self.phase = 2
          
          from py2play.idler import IDLER
          
          self.traveling = soya.FixTraveling(None, soya.Vector(IDLER.camera, 0.0, 0.0, -1.0))
          #self.traveling = soya.ThirdPersonTraveling(self)
          #self.traveling.distance = 400.0
          IDLER.camera.add_traveling(self.traveling)
          
      self.p.__init__(self.target, 0.0, 0.0, -4.0)
      
      #from py2play.idler  import IDLER
      #if hasattr(IDLER.camera.traveling, "distance") and (IDLER.camera.traveling.distance > 5.0): IDLER.camera.traveling.distance -= 1.0
      
    elif self.phase == 2:
      if self.y < self.drop_y:
        self.phase = 3

        from py2play.idler  import IDLER
        IDLER.camera.remove_traveling(self.traveling)
        
        for character in self.parent.characters:
          if self.distance_to(character) < 3.0:
            character.die()
            break
        else:
          sound.play("explose-3.wav", self)
          
          explode = slune_character.Explosion(self.parent)
          explode.move(self)
          explode.life = 30
          
      #from py2play.idler  import IDLER
      #if hasattr(IDLER.camera.traveling, "distance") and (IDLER.camera.traveling.distance < 20.0): IDLER.camera.traveling.distance += 1.0
      
      
    elif self.phase == 3:
      if self.y > self.drop_y + 10.0:
        self.phase = 1
        
        
      #from py2play.idler  import IDLER
      #if hasattr(IDLER.camera.traveling, "distance") and (IDLER.camera.traveling.distance > 5.0): IDLER.camera.traveling.distance -= 1.0
      
  def advance_time(self, proportion):
    soya.World.advance_time(self, proportion)
    
    if self.play == 0: return
    
    if   self.phase == 1:
      self.v.set_start_end(self, self.p)
      self.v.y = 0
      self.v.set_length(self.speed * proportion)
      self.look_at(self.v)
      self += self.v
      
    elif self.phase == 2:
      self.y -= self.speed * proportion
      
    elif self.phase == 3:
      self.y += self.speed * proportion
      



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya, soya.opengl as soyaopengl, soya.widget as widget
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
import slune.character as slune_character
import slune.controler
import slune.globdef as globdef, slune.sound as sound, slune.level as slune_level

def init_race_game(level):
  """Hides all flags except the first."""
  
  # Copy them in level, since the different player may have different config.
  level.race_nb_laps      = globdef.RACE_NB_LAPS
  level.race_nb_opponents = globdef.RACE_NB_OPPONENTS
  
  level.ranks = []
  
  level.flags = [] # Indexes all flags
  
  for item in level:
    if isinstance(item, Flag):
      item.visible = (item.name == "flag_0")
      level.flags.append(item)
      
  level.flags.sort(lambda f1, f2: cmp(f1.flag_id, f2.flag_id))
  
  
def start_race_game(level, add_opponent1 = 1):
  """Inits time property and shows the race time label."""
  for i in range(level.race_nb_opponents - len(level.characters)):
    Opponent().set_level(level, 1)
    
  Checker(level)
  
  for character in level.characters:
    if isinstance(character, slune_character.Competitor): init_race_opponent(character)
    
def init_race_opponent(character, laps_label = 1, minimap = 1):
  character.next_flag_id = 0
  character.nb_laps   = 0
  character.lap_start = time.time()
  character.best_lap  = 9999.0
  if character.played and character.player.active:
    from py2play.idler import IDLER
    if laps_label: IDLER.no_blackbands_group.insert(0, LapsLabel(character))
    if minimap:    IDLER.no_blackbands_group.insert(0, MiniMap(character.level))

def end_race_game(level):
  """Hides the race time label."""
  for widget in soya.root_widget.children:
    if isinstance(widget, LapsLabel):
      soya.root_widget.remove(widget)
      break
    
class Checker(soya.Volume):
  def begin_round(self):
    from py2play.idler import IDLER
    
    level = self.parent
    
    for character in level.characters:
      if isinstance(character, slune_character.Competitor):
        flag = level.flags[character.next_flag_id]
        if character.distance_to(flag) < (flag.radius + character.radius):
          
          #character.life = min(1.0, character.life + 0.1)
          
          if character.next_flag_id + 1 < len(self.parent.flags):
            next_flag_id = character.next_flag_id + 1
            next_flag = self.parent.flags[next_flag_id]
          else:
            next_flag_id = 0
            next_flag = self.parent.flags[0]
            character.nb_laps += 1
            now = time.time()
            this_lap = now - character.lap_start
            if this_lap < character.best_lap: character.best_lap = this_lap
            character.lap_start = now
            
            if character.nb_laps >= character.level.race_nb_laps:
              firework = soya.FlagFirework(level, nb_particles = 6)
              firework.move(flag)
              
              character.next_flag_id = 0
              
              level.ranks.append(character)

              if len(level.ranks) == 1: # Winner
                IDLER.level_completed(character, 1, _("__race_winner__"))
              else:
                IDLER.level_completed(character, 0, _("__race_looser__") % len(level.ranks))
                
              from py2play.player import CURRENT_PLAYER
              if not character is CURRENT_PLAYER.character:
                IDLER.message(_("__race_arrival__") % (character.displayname(), len(level.ranks)))
                
              if not character.played:
                character.level.remove_character(character)
              return
            
          character.next_flag_id = next_flag_id
          
          sound.play("flag-1.wav", character)
          
          if character.played and character.player.active:
            flag.visible = 0
            next_flag.visible = 1
            
            firework = soya.Smoke(level, nb_particles = 10, removable = 1)
            firework.move(flag)
            firework.regenerate()
            
    # check if every one is killer, except the current player.
    from py2play.player import CURRENT_PLAYER
    
    if (len(self.parent.characters) == 1) and (self.parent.characters[0] is CURRENT_PLAYER.character):
      self.parent.ranks.append(CURRENT_PLAYER.character)
      if len(self.parent.ranks) == 1:
        IDLER.level_completed(CURRENT_PLAYER.character, 1, message = _("__race_winner__"))
      else:
        IDLER.level_completed(CURRENT_PLAYER.character, 1, message = _("__race_looser__") % len(self.parent.ranks))
        
      self.parent.remove(self)
  
  
class Flag(soya.Bonus):
  radius = 3.5
  
  def __init__(self, parent = None):
    if parent:
      parent.max_laps = 2
      
      nb = 0
      for item in parent:
        if isinstance(item, Flag): nb += 1
        
      soya.Bonus.__init__(self, parent, soya.Material.get("flag1"), soya.Material.get("halo"))
      
      self.flag_id = nb
      
    else:
      soya.Bonus.__init__(self, parent, soya.Material.get("flag1"), soya.Material.get("halo"))
      
    self.color = (1.0, 1.0, 0.20000000298023224, 1.0)
    
  def get_flag_id(self):
    try: return int(self.name[5:])
    except ValueError: return 0
  def set_flag_id(self, id): self.name = "flag_%s" % id
  flag_id = property(get_flag_id, set_flag_id)


class MiniMap(widget.Widget):
  def __init__(self, level):
    self.level = level
    
    self.top = 10
    self.width  = 200
    self.height = 200
    self.left = soya.root_widget.width - self.width - 40
    
    flags = [x for x in level if isinstance(x, Flag)]
    flags.sort(lambda f1, f2: cmp(f1.flag_id, f2.flag_id))
    min_x = min_y =  10e1000
    max_x = max_y = -10e1000
    for flag in flags:
      if flag.x < min_x: min_x = flag.x
      if flag.x > max_x: max_x = flag.x
      if flag.z < min_y: min_y = flag.z
      if flag.z > max_y: max_y = flag.z
    f_x = self.width  / (max_x - min_x)
    f_y = self.height / (max_y - min_y)
    if f_x < f_y:
      f = f_x
      left = self.left
    else:
      f = f_y
      left = self.left + ((max_y - min_y) - (max_x - min_x)) * f
    top = self.top
    
    self.f     = f
    self.cleft = left
    self.ctop  = top
    self.min_x = min_x
    self.min_y = min_y
    
    self.flags_pos = []
    self._calllist = soyaopengl.glGenList()
    soyaopengl.glNewList(self._calllist)
    soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    soyaopengl.glBegin(soyaopengl.GL_LINE_LOOP)
    for flag in flags:
      self.flags_pos.append((
        left + f * (flag.x - min_x),
        top  + f * (flag.z - min_y),
        ))
      soyaopengl.glVertex2f(*self.flags_pos[-1])
    soyaopengl.glEnd()
    x, y = self.flags_pos[-1]
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    soyaopengl.glVertex2f(x - 4.0, y - 4.0)
    soyaopengl.glVertex2f(x - 4.0, y + 4.0)
    soyaopengl.glVertex2f(x + 4.0, y + 4.0)
    soyaopengl.glVertex2f(x + 4.0, y - 4.0)
    soyaopengl.glEnd()
    
    soyaopengl.glEndList()
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.left = parent_width - self.width - 40
    self.screen_width  = parent_width
    self.screen_height = parent_height
    
  def render(self):
    from py2play.player import CURRENT_PLAYER
    from py2play.idler import IDLER
    ca = IDLER.camera
    x, y, z = ca.transform_point(0.0, 0.0, 0.0, self.level.flags[CURRENT_PLAYER.character.next_flag_id])
    angle = math.atan(x / z)
    if z < 0.0:
      vx = -math.sin(angle)
      vy = -math.cos(angle)
    else:
      vx = math.sin(angle)
      vy = math.cos(angle)
    
    r = self.screen_width / self.screen_height
    if r > abs(vx / vy):
      if vy < 0.0:
        my = -self.screen_height / 2.0
        mx = vx * (my / vy) * r
      else:
        my = self.screen_height / 2.0
        mx = vx * (my / vy) * r
    else:
      if vx < 0.0:
        mx = -self.screen_width / 2.0
        my = vy * (mx / vx) / r
      else:
        mx = self.screen_width / 2.0
        my = vy * (mx / vx) / r
        
    x = mx + self.screen_width  / 2.0
    y = my + self.screen_height / 2.0
    
    x1 = x + (+ my - mx) * 0.06
    x2 = x + (- my - mx) * 0.06
    y1 = y + (- mx - my) * 0.06
    y2 = y + (+ mx - my) * 0.06

    if globdef.QUALITY > 1:
      soyaopengl.glEnable(soyaopengl.GL_BLEND)
      soyaopengl.glColor4f(0.0, 0.0, 0.0, 0.5)
      soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
      soyaopengl.glVertex2f(x  - 9.0, y  + 9.0)
      soyaopengl.glVertex2f(x1 - 9.0, y1 + 9.0)
      soyaopengl.glVertex2f(x2 - 9.0, y2 + 9.0)
      soyaopengl.glEnd()
      soyaopengl.glDisable(soyaopengl.GL_BLEND)
    
    soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    soyaopengl.glBegin(soyaopengl.GL_TRIANGLES)
    soyaopengl.glVertex2f(x, y)
    soyaopengl.glVertex2f(x1, y1)
    soyaopengl.glVertex2f(x2, y2)
    soyaopengl.glEnd()
    
    soyaopengl.glCallList(self._calllist)
    for character in self.level.characters:
      if isinstance(character, slune_character.Competitor):
        x = self.cleft + self.f * (character.x - self.min_x)
        y = self.ctop  + self.f * (character.z - self.min_y)
        
        material = soya.Material.get("head_" + character.perso_name + "_1")
        material.activate()
        soyaopengl.glEnable(soyaopengl.GL_BLEND)
        soyaopengl.glBegin(soyaopengl.GL_QUADS)
        soyaopengl.glTexCoord2f(0.0, 0.0); soyaopengl.glVertex2f(x - 20.0, y - 20.0)
        soyaopengl.glTexCoord2f(0.0, 1.0); soyaopengl.glVertex2f(x - 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 1.0); soyaopengl.glVertex2f(x + 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 0.0); soyaopengl.glVertex2f(x + 20.0, y - 20.0)
        soyaopengl.glEnd()
        soyaopengl.glDisable(soyaopengl.GL_BLEND)
        soya.DEFAULT_MATERIAL.activate()

#         if   character.perso_name == "tux":
#           material = soya.Material.get("tux1"); material.activate()
#           soyaopengl.glBegin(soyaopengl.GL_QUADS)
#           soyaopengl.glTexCoord2f(0.6, 0.0); soyaopengl.glVertex2f(x - 10.0, y - 10.0)
#           soyaopengl.glTexCoord2f(0.6, 1.0); soyaopengl.glVertex2f(x - 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0, 1.0); soyaopengl.glVertex2f(x + 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0, 0.0); soyaopengl.glVertex2f(x + 10.0, y - 10.0)
#           soyaopengl.glEnd()
#         elif character.perso_name == "gnu":
#           material = soya.Material.get("gnu"); material.activate()
#           soyaopengl.glBegin(soyaopengl.GL_QUADS)
#           soyaopengl.glTexCoord2f(0.52, 0.4); soyaopengl.glVertex2f(x - 10.0, y - 10.0)
#           soyaopengl.glTexCoord2f(0.52, 1.0); soyaopengl.glVertex2f(x - 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0 , 1.0); soyaopengl.glVertex2f(x + 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0 , 0.4); soyaopengl.glVertex2f(x + 10.0, y - 10.0)
#           soyaopengl.glEnd()
#         elif character.perso_name == "shark":
#           material = soya.Material.get("shark1"); material.activate()
#           soyaopengl.glBegin(soyaopengl.GL_QUADS)
#           soyaopengl.glTexCoord2f(0.46, 0.48); soyaopengl.glVertex2f(x - 10.0, y - 10.0)
#           soyaopengl.glTexCoord2f(0.46, 1.0 ); soyaopengl.glVertex2f(x - 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0 , 1.0 ); soyaopengl.glVertex2f(x + 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0 , 0.48); soyaopengl.glVertex2f(x + 10.0, y - 10.0)
#           soyaopengl.glEnd()
#         elif character.perso_name == "python":
#           material = soya.Material.get("python1"); material.activate()
#           soyaopengl.glBegin(soyaopengl.GL_QUADS)
#           soyaopengl.glTexCoord2f(1.0 , 0.36); soyaopengl.glVertex2f(x - 10.0, y - 10.0)
#           soyaopengl.glTexCoord2f(0.56, 0.36); soyaopengl.glVertex2f(x - 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(0.56, 1.0 ); soyaopengl.glVertex2f(x + 10.0, y + 10.0)
#           soyaopengl.glTexCoord2f(1.0 , 1.0 ); soyaopengl.glVertex2f(x + 10.0, y - 10.0)
#           soyaopengl.glEnd()
        
        
class LapsLabel(widget.Label):
  def __init__(self, character):
    self.character = character
    self.format = str(_("__racelapsformat__"))
    widget.Label.__init__(self)
    
    self.left = 10
    self.width  = 400
    self.height = 100
    self.old_nb_laps = -1
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.top = parent_top + soya.root_widget.height - 43
    
  def widget_begin_round(self):
    if self.old_nb_laps != self.character.nb_laps:
      self.text = self.format % (self.character.nb_laps)
      self.old_nb_laps = self.character.nb_laps
    

class RaceChromosom:
  def __init__(self, max_speed, innitiative, will, agressivity, turbosity):
    self.max_speed   = max_speed
    self.innitiative = innitiative # Lower is better
    self.agressivity = agressivity
    self.will        = will
    self.turbosity   = turbosity
    
  def mutate(self, random):
    for attr, value in self.__dict__.iteritems():
      if attr != "max_speed":
        if random.random() < 0.4: 
          setattr(self, attr, value * (0.5 + random.random()))
      
    
  
# CHROMOSOMS = {
#   ("tux", 0): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.2 , 80.0, 30.0, 0.5],
#   ("tux", 1): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.55, 30.0, 50.0, 0.5],
#   ("tux", 2): [0.23161940934461928, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 1.0 , 20.0, 50.0, 0.6],
  
#   ("gnu", 0): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.1, 50.0, 50.0, 0.7],
#   ("gnu", 1): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.5 , 15.0, 80.0, 0.8],
#   ("gnu", 2): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 1.0 , 10.0, 99.0, 0.8],
  
#   ("shark", 0): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.1 , 50.0, 50.0, 0.4],
#   ("shark", 1): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 0.5 , 15.0, 50.0, 0.3],
#   ("shark", 2): [0.27955060118474129, 0.94713263528551006, 0.0076398049940541654, 0.46242007806615826, 22.758284696585228, 0.07628649333395944, 0.0051247277708940846, 5.6391784712330182, 1.0 , 10.0, 60.0, 0.2],
#   }
CHROMOSOMS = {
  ("tux",    0): RaceChromosom(0.2,  80.0,  30.0, 0.5, -7.0),
  ("tux",    1): RaceChromosom(0.6,  30.0,  50.0, 0.5, -1.0),
  ("tux",    2): RaceChromosom(1.1,  20.0,  50.0, 0.6,  0.5),
  
  ("gnu",    0): RaceChromosom(0.1,  50.0,  60.0, 0.7, -2.0),
  ("gnu",    1): RaceChromosom(0.5,  15.0,  90.0, 0.8,  0.7),
  ("gnu",    2): RaceChromosom(1.0,  10.0, 150.0, 0.9,  1.0),
  
  ("shark",  0): RaceChromosom(0.1,  40.0,  50.0, 0.3, -9.0),
  ("shark",  1): RaceChromosom(0.5,  15.0,  50.0, 0.3, -4.0),
  ("shark",  2): RaceChromosom(1.0,  10.0,  60.0, 0.2,  0.0),
  
  ("python", 0): RaceChromosom(0.2,  70.0,  30.0, 0.4, -1.0),
  ("python", 1): RaceChromosom(0.55, 25.0,  40.0, 0.4,  1.3),
  ("python", 2): RaceChromosom(1.0,  15.0,  50.0, 0.4,  2.0),
  }


class Opponent(slune_character.NonPlayerCharacter, slune_character.Competitor):
  def __init__(self, parent = None, chromosom = None):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    self.chromosom = chromosom
    
  def init(self):
    if not self.perso_name:
      persos = list(globdef.CHARACTERS)
      self.random.shuffle(persos)
      for perso in persos:
        perso = perso.lower()
        for c in self.level:
          if (hasattr(c, "perso_name")) and (c.perso_name == perso): break
        else:
          self.perso_name = perso
          break
      else: self.perso_name = persos[0].lower()
    self.set_perso(self.perso_name)
    
    if self.vehicle == 1: self.set_vehicle(self.random.choice(globdef.VEHICLES.values()))
    
  def IAControler(self):
    c = self.chromosom = self.chromosom or CHROMOSOMS[self.perso_name, self.level.difficulty]
    self.max_speed = c.max_speed
    
    self.speed.y = 100.0 # Avoid null speed (null speed => jumping). The speed will be overrided later.
    
    while 1:
      for i in range(int(c.innitiative)):
        yield slune.controler.goto(self, self.level.flags[self.next_flag_id], turbosity = c.turbosity + 2.0 * self.life)
        
      # Check boosting
      p = [p for p in self.level.characters if p.played]
      if p:
        if len(p) == 1: p = p[0]
        else:           p = self.random.choice(p)
        delta = (self.nb_laps * len(self.level.flags) + self.next_flag_id) - (p.nb_laps * len(p.level.flags) + p.next_flag_id)
        if   (delta >  1.0) and p.level.difficulty < 2: self.max_speed = self.max_speed * 0.9
        elif (delta < -1.0) and p.level.difficulty > 0: self.max_speed = 1.1
        elif delta > 0.0: self.max_speed = c.max_speed
      else: self.max_speed = c.max_speed
      
      yield slune.controler.CheckBonus(self, self.chromosom.agressivity)
      
      if self.random.random() < c.agressivity: # Attack
        for competitor in self.level.characters:
          if not competitor is self:
            px, py, pz = self.transform_point(competitor.x, competitor.y, competitor.z, competitor.parent)
            if (-5.0 < pz < 0.0) and (-2.0 < py < 2.0) and (-5.0 < px < 5.0):
              yield slune.controler.AttackRun(self, competitor, c.will)
              break
            
            elif competitor.nb_laps >= self.nb_laps + 2: # Too late to win the race, try to kill everybody
              while 1:
                dist = 1000000.0
                best = None
                for competitor in self.level.characters:
                  d = self.distance_to(competitor) - 20.0 * competitor.nb_laps
                  if d < dist:
                    dist = d
                    best = competitor
                yield slune.controler.AttackRun(self, best, c.will)
                
          
      else: # Dodge
        for competitor in self.level.characters:
          if not competitor is self:
            px, py, pz = competitor.transform_point(self.x, self.y, self.z, self.parent)
            if (-5.0 < pz < 0.0) and (-2.0 < py < 2.0) and (-5.0 < px < 5.0):
              h = self.random.random()
              if h < 0.3:
                for i in range(10): yield slune_character.Action(slune_character.ACTION_WAIT, -1.0, self.max_speed)
              elif h < 0.6:
                for i in range(10): yield slune_character.Action(slune_character.ACTION_WAIT,  1.0, self.max_speed)
              else:
                jumping = 1
                
                yield slune_character.Action(slune_character.ACTION_JUMP, 0.0, self.max_speed)
                for i in range(2): yield slune_character.Action(slune_character.ACTION_ROTDOWN, 0.0, self.max_speed)
              break



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This script is used for genetically optimizing racing IA.
# It considers only 1 character at a time; see race_algogen2 for a version with 2 IA.

from __future__ import generators

RENDER = 0
PERSO  = "tux"
LEVEL  = "crashfunkel-race-1"
#LEVEL  = "crashfunkel-race-2"
#LEVEL  = "canion-race-1"
#LEVEL  = "canion-race-3"
import sys

i = 1
while i < len(sys.argv):
  arg = sys.argv[i]
  if   arg == "--render": RENDER = 1
  
  elif arg == "--perso":
    i += 1
    PERSO = sys.argv[i]
  
  elif arg == "--level":
    i += 1
    LEVEL = sys.argv[i]
  
  elif arg == "--help":
    print """Optimize IA, by running 1 IA against the clock.
The resulting params are printed on stdout, and should be manually copied in race.py.

Options:
  --render         Render the 3D, so as you can see something (else no rendering is performed, for speedup).
  --perso <perso>  Choose the character IA to optimize (e.g. tux, gnu,...).
  --level <level>  Choose the level. Should be a racing level."""
    sys.exit()
    
  else:
    print "Unknown options %s !" % arg
    sys.exit(1)
    
  i += 1

IA_PARAMS = [
"slune.controler.GOTO_X_2_TURN",
"slune.controler.GOTO_X_2_SPEED",
"slune.controler.GOTO_Y_2_SPEED",
"slune.controler.GOTO_Z_2_SPEED",
"slune.controler.GOTO_NOGROUND_2_SPEED",
"slune.controler.GOTO_Y_2_JUMP",
"slune.controler.GOTO_Z_2_JUMP",
"slune.controler.GOTO_NOGROUND_2_JUMP",
  ]

import random
import soya, soya.soya3d as soya3d, soya.game.level as soya_game_level, soya.game.camera as soya_camera
import py2play.level
import slune.race as race, slune.level, slune.character, slune.player, slune.controler

class Idler(slune.level.Idler):
  def __init__(self):
    slune.level.Idler.__init__(self)
    self.next_flag_name = ""
    
  def idle(self):
    race.init_race_opponent(OPPONENT)
    
    for nb_round in range(3300):
      self.begin_round()
      self.advance_time(1.0)
      self.end_round()
      
      if self.next_round_tasks:
        for task in self.next_round_tasks: task()
        self.next_round_tasks = []
        
      self.render()
      
      if OPPONENT.best_lap != 9999.0: break
      
    return nb_round
  
  def render(self):
    if RENDER: slune.level.Idler.render(self)
    
    if hasattr(OPPONENT, "next_flag_name"):
      if self.next_flag_name != OPPONENT.next_flag_name:
        self.next_flag_name = OPPONENT.next_flag_name
        print "Next flag: ", self.next_flag_name
        
  
soya.init(width = 640, height = 480, fullscreen = 0)

idler = Idler()

import py2play.player
py2play.player.CURRENT_PLAYER = py2play.player.Player("fake") # HACK !
py2play.player.CURRENT_PLAYER.character = None

level = py2play.level.CREATE("level-" + LEVEL)
level.set_active(1)
race.init_race_game (level)
race.start_race_game(level)

for character in level.characters[:]: level.remove_character(character)

class Opponent(race.Opponent):
  def __init__(self, *args, **kargs):
    race.Opponent.__init__(self, *args, **kargs)
    
    self.value = 0
    
OPPONENT = None
def test(opponent):
  global OPPONENT
  OPPONENT = opponent
  OPPONENT.perso_name = PERSO
  
  idler.camera.add_traveling(soya_camera.ThirdPersonTraveling(OPPONENT.camera_target))
  
  OPPONENT.set_level(level, 1)
  OPPONENT.set_vehicle(1)
  
  for param in IA_PARAMS: exec """%s = %s""" % (param, OPPONENT.params[param])
  
  value = idler.idle()
  print value
  
  OPPONENT.parent.remove_character(OPPONENT)
  
  return value

def value_sorter(a, b): return cmp(a.value, b.value)

def mutate(value):
  if random.random() < 0.8: return value
  if not value: return 0.02
  return value + (random.random() - 0.5) * value

def make_love(father, mother):
  o = Opponent(chromosom = father.chromosom)
  o.params = dict(map(lambda param: (param, mutate(random.choice([father.params[param], mother.params[param]]))), IA_PARAMS))
  return o

OPPONENTS = [Opponent(chromosom = race.CHROMOSOMS[(PERSO, 2)]), Opponent(chromosom = race.CHROMOSOMS[(PERSO, 2)])]
for opponent in OPPONENTS:
  opponent.params = dict(map(lambda param: (param, eval(param)), IA_PARAMS))
  
while 1:
  for i in range(4):
    OPPONENTS.append(make_love(random.choice(OPPONENTS), random.choice(OPPONENTS)))
    
  for opponent in OPPONENTS:
    if not opponent.value: opponent.value = test(opponent)
      
  OPPONENTS.sort(value_sorter)
  OPPONENTS = OPPONENTS[:4]

  print
  for opponent in OPPONENTS:
    for k, v in opponent.params.iteritems():
      print k, "=", v
    print "=>", opponent.value
    print



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Tkinter, socket, os, os.path
import slune.globdef as globdef, soya, soya.model, soya.soya3d as soya3d

class MainScreen(Tkinter.Tk):
  def __init__(self):
    globdef.GUI = 1 # Using GUI !
    
    Tkinter.Tk.__init__(self)
    self.wm_title("Slune")
    self.option_add("*Entry.relief"      , "flat")
    self.option_add("*Entry.background"  , "#FFFFFF")
    self.option_add("*Text.relief"       , "flat")
    self.option_add("*Text.background"   , "#FFFFFF")
    self.option_add("*Listbox.relief"    , "flat")
    self.option_add("*Listbox.background", "#FFFFFF")
    
    self.image = Tkinter.PhotoImage(master = self, file = os.path.join(soya.model.Image.PATH, "slune-splash.pgm"))
    Tkinter.Label(self, image = self.image, borderwidth = 0).grid(column = 0, row = 0, columnspan = 2)
    
    self.columnconfigure(0, weight = 1)
    self.columnconfigure(1, weight = 0)
    
    self.rowconfigure(0, weight = 1)
    self.rowconfigure(1, minsize = 50)
    self.rowconfigure(2, minsize = 50)
    
    Tkinter.Button(self, text = _("Start new game"   ), command = self.on_new_game ).grid(column = 0, row = 1, sticky = "NSEW")
    Tkinter.Button(self, text = _("Join network game"), command = self.on_join_game).grid(column = 0, row = 2, sticky = "NSEW")
    Tkinter.Button(self, text = _("Options"          ), command = self.on_option   ).grid(column = 0, row = 3, sticky = "NSEW")
    Tkinter.Button(self, text = _("References"       ), command = self.on_info     ).grid(column = 0, row = 4, sticky = "NSEW")
    Tkinter.Button(self, text = _("Credits"          ), command = self.on_credits  ).grid(column = 0, row = 5, sticky = "NSEW")
    Tkinter.Button(self, text = _("Quit"             ), command = self.on_quit     ).grid(column = 0, row = 6, sticky = "NSEW")
    self.wm_protocol("WM_DELETE_WINDOW", self.on_quit)
    
    new_game_options = Tkinter.Frame(self)
    new_game_options.grid(column = 1, row = 1, sticky="NSEW")
    new_game_options.columnconfigure(0, weight = 1)
    new_game_options.columnconfigure(1, weight = 0)
    new_game_options.columnconfigure(2, weight = 0)
    new_game_options.columnconfigure(3, weight = 0)
    
    Tkinter.Label(new_game_options, anchor="e", text = _("Choose level")).grid(column = 0, row = 0, sticky = "NSEW")
    self.level = Tkinter.StringVar()
    self.level.set("mission-1")
    
    levels = map(lambda level_name: level_name[6:], filter(lambda world_name: world_name and world_name.startswith("level-"), soya.soya3d.World.availables()))
    Tkinter.OptionMenu(new_game_options, self.level, *levels).grid(column = 1, row = 0, columnspan = 3, sticky = "NSEW")
    
    Tkinter.Label(new_game_options, anchor="e", text = _("You are")).grid(column = 0, row = 1, sticky="NSEW")
    
    var = Tkinter.StringVar()
    var.set(str(globdef.NAME))
    self.name = Tkinter.Entry(new_game_options, textvariable = var, width = 10, justify = "right")
    self.name.grid(column = 1, row = 1, sticky="NSEW")
    
    Tkinter.Label(new_game_options, text = "@" + socket.gethostname() + ":").grid(column = 2, row = 1, sticky = "NSEW")
    
    self.port = Tkinter.StringVar()
    self.port.set(str(globdef.PORT))
    Tkinter.Entry(new_game_options, textvariable = self.port, width = 5).grid(column = 3, row = 1)
    
    join_game_options = Tkinter.Frame(self)
    join_game_options.grid(column = 1, row = 2, sticky="NSEW")
    join_game_options.columnconfigure(0, weight = 1)
    join_game_options.columnconfigure(1, weight = 1)
    join_game_options.columnconfigure(2, weight = 0)
    join_game_options.columnconfigure(3, weight = 0)
    
    Tkinter.Label(join_game_options, anchor="e", text = _("Join at (host or IP:port)")).grid(column = 0, row = 1, sticky="NSEW")
    
    self.parrain_host = Tkinter.StringVar()
    self.parrain_host.set(str(globdef.PARRAIN_HOST))
    Tkinter.Entry(join_game_options, textvariable = self.parrain_host, width = 18, justify = "right").grid(column = 1, row = 1, sticky="NSEW")
    
    Tkinter.Label(join_game_options, text = ":").grid(column = 2, row = 1, sticky="NSEW")
    
    self.parrain_port = Tkinter.StringVar()
    self.parrain_port.set(globdef.PARRAIN_PORT)
    Tkinter.Entry(join_game_options, textvariable = self.parrain_port, width = 5).grid(column = 3, row = 1, sticky="NSEW")
    
    Tkinter.Label(self, text = _("Slune version %s") % globdef.VERSION).grid(column = 1, row = 6, sticky="NSEW")
    
    character_options = Tkinter.Frame(self)
    character_options.grid(column = 1, row = 3, sticky="NSEW")
    character_options.columnconfigure(0, weight = 0)
    character_options.columnconfigure(1, weight = 1)
    character_options.columnconfigure(2, weight = 0)
    character_options.columnconfigure(3, weight = 1)
    
    Tkinter.Label(character_options, anchor = "e", text = _("Character")).grid(column = 0, row = 0, sticky="NSEW")
    var = Tkinter.StringVar()
    var.set(globdef.CHARACTER)
    Tkinter.OptionMenu(character_options, var, command = self.on_character, *globdef.CHARACTERS).grid(column = 1, row = 0, sticky = "NSEW")
    
    Tkinter.Label(character_options, anchor = "e", text = _("Vehicle")).grid(column = 2, row = 0, sticky="NSEW")
    var = Tkinter.StringVar()
    for text, level in globdef.VEHICLES.items():
      if level == globdef.VEHICLE:
        var.set(text)
        break
    Tkinter.OptionMenu(character_options, var, command = self.on_vehicle, *globdef.VEHICLES.keys()).grid(column = 3, row = 0, sticky = "NSEW")
    
    difficulty_options = Tkinter.Frame(self)
    difficulty_options.grid(column = 1, row = 4, sticky="NSEW")
    difficulty_options.columnconfigure(0, weight = 0)
    difficulty_options.columnconfigure(1, weight = 1)
    
    Tkinter.Label(difficulty_options, anchor = "e", text = _("Difficulty level")).grid(column = 0, row = 0, sticky="NSEW")
    var = Tkinter.StringVar()
    for text, level in globdef.DIFFICULTIES.items():
      if level == globdef.DIFFICULTY:
        var.set(text)
        break
    Tkinter.OptionMenu(difficulty_options, var, _("Newbie"), _("Hacker"), _("Guru"), command = self.on_difficulty).grid(column = 1, row = 0, sticky = "NSEW")
    
    x = self.winfo_screenwidth () / 2 - 385
    y = self.winfo_screenheight() / 2 - 300
    self.wm_geometry("+%s+%s" % (x, y))
    
  def on_new_game(self):
    self.read_params()
    
    self.lower()
    
    import slune.player
    soya.init(width = globdef.SCREEN_WIDTH, height = globdef.SCREEN_HEIGHT)
    soya.set_quality(globdef.QUALITY)
    if globdef.FULLSCREEN: soya.toggle_fullscreen()
    r = slune.player.start_new_game("level-" + self.level.get())
    if globdef.FULLSCREEN: soya.toggle_fullscreen()
    
    if r:
      if os.path.exists(os.path.join(soya3d.World.PATH, "level-" + r[0] + ".data")):
        self.level.set(r[0])
        
    self.tkraise()
    
  def on_join_game(self):
    self.read_params()
    if (globdef.PARRAIN_HOST == "") and (globdef.PORT == globdef.PARRAIN_PORT):
      import tkMessageBox
      tkMessageBox.showerror(_("Network error!"), _("__networkbadconfig__"))
      return
    
    self.lower()
    
    import slune.player
    soya.init(width = globdef.SCREEN_WIDTH, height = globdef.SCREEN_HEIGHT)
    soya.set_quality(globdef.QUALITY)
    if globdef.FULLSCREEN: soya.toggle_fullscreen()
    slune.player.join_network_game()
    if globdef.FULLSCREEN: soya.toggle_fullscreen()
    
    #soya.set_video(640, 480, 0, 0)
    #soya.minimize()
    
    self.tkraise()
    
  def on_option(self):
    self.read_params()
    
    import slune.gui_option
    slune.gui_option.OptionScreen()
    
  def on_info(self): info()
    
  
  def on_credits(self):
    pass
  
  def on_quit(self):
    self.read_params()
    globdef.generate_dot_slune()
    import sys; sys.exit()
    
  def read_params(self):
    globdef.NAME         =     self.name        .get()
    globdef.PORT         = int(self.port        .get())
    globdef.PARRAIN_HOST =     self.parrain_host.get()
    globdef.PARRAIN_PORT = int(self.parrain_port.get())
    
  def on_difficulty(self, difficulty): globdef.DIFFICULTY = globdef.DIFFICULTIES[difficulty]
  
  def on_character (self, character ): globdef.CHARACTER  = character
  
  def on_vehicle   (self, vehicle   ): globdef.VEHICLE    = globdef.VEHICLES[vehicle]


def info():
  f = Tkinter.Toplevel()
  f.wm_title(_("Slune info sources"))
  t = Tkinter.Text(f, background="white")
  t.tag_configure("link", foreground="blue", font = "Courier -11 underline")
  t["font"] = "Courier -16"
  t.pack()
  text = _("__info__")
  for line in text.split("\n"):
    for word in line.split(" "):
      if word.startswith("http://"):
        t.insert("end", word, ("link", word))
        def browse(event, url = word):
          import webbrowser
          webbrowser.open(url)
        t.tag_bind(word, "<Button-1>", browse)
      else: t.insert("end", word)
      t.insert("end", " ")
    t.insert("end", "\n")



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import soya, soyasdlconst
import editobj, editobj.editor as editor, editobj.custom as custom
import soya.editor, soya.editor.world
from soya import Point, Vector

import slune.level, slune.race, slune.fight, slune.character

custom.EVAL_ENV.update({
  "race"  : slune.race,
  "fight" : slune.fight,
  })

custom.register_attr("script"               , editor.TextEditor)
custom.register_attr("init_script"          , editor.TextEditor)
custom.register_attr("init_character_script", editor.TextEditor)
custom.register_attr("start_script"         , editor.TextEditor)
custom.register_attr("end_script"           , editor.TextEditor)

slune.race.Flag                .__clickmanager__ = soya.editor.world.MoveClickManager
slune.level.LifeBonus          .__clickmanager__ = soya.editor.world.MoveClickManager
slune.level.FlameThrowerBonus  .__clickmanager__ = soya.editor.world.MoveClickManager
slune.level.SphericalScript    .__clickmanager__ = soya.editor.world.MoveClickManager
slune.character.Pushable       .__clickmanager__ = soya.editor.world.MoveClickManager
slune.character.TakeableVehicle.__clickmanager__ = soya.editor.world.MoveClickManager

_LAST_BARRIER = None
_LAST_BARRIER_TYPE = 0
_BARIER_TYPES = [slune.character.Barrier, slune.character.Barrier2, slune.character.TreeTrunc, slune.character.Quille, slune.character.Ball]

def _add_flag              (root, current): slune.race.Flag(root)
def _add_life_bonus        (root, current): slune.level.LifeBonus(root)
def _add_flamethrower_bonus(root, current): slune.level.FlameThrowerBonus(root)
def _add_sphericalscript   (root, current): slune.level.SphericalScript(root)
def _add_takeablevehicle   (root, current): slune.character.TakeableVehicle(root)
def _add_barrier           (root, current):
  global _LAST_BARRIER
  _LAST_BARRIER = _BARIER_TYPES[_LAST_BARRIER_TYPE](root)
def _change_last_barrier(root, current):
  global _LAST_BARRIER_TYPE
  if _LAST_BARRIER_TYPE < len(_BARIER_TYPES) - 1: _LAST_BARRIER_TYPE += 1
  else:                                           _LAST_BARRIER_TYPE  = 0
  _BARIER_TYPES[_LAST_BARRIER_TYPE](root, _LAST_BARRIER)

def _put_down(root, current):
  cursor = root.parent.parent["cursor"]
  
  result = root.raypick(cursor, Vector(root, 0.0, -1.0, 0.0), -1.0, 2)
  
  if result:
    p, v = result
    cursor.add_vector(cursor >> p)
    cursor.button_pressed(1)
    cursor.button_released(1)
    
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_d] = _put_down
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_f] = _add_flag
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_s] = _add_sphericalscript
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_b] = _add_barrier
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_n] = _change_last_barrier
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_l] = _add_life_bonus
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_p] = _add_flamethrower_bonus
soya.editor.world.KEY_BINDINGS[soyasdlconst.K_v] = _add_takeablevehicle



# Slune
# Copyright (C) 2002-2003 Bertrand LAMY, Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, os, os.path, random, math

import soya, soya.opengl as soyaopengl, soya.sdlconst as soyasdlconst
import soya.widget   as widget

import slune.globdef as globdef

current_menu = None
old_menus = []

def MainScreen():

  #------------------------------#
  # internal classes for widgets #
  #------------------------------#

  def draw_window(width, height, color = (0.5, 0.5, 0.5, 0.75), border_size = 20):
    x = int(0.5 * (soya.get_screen_width()  - width))
    y = int(0.5 * (soya.get_screen_height() - height))
    w = x + width
    h = y + height
    bx = x - border_size
    by = y - border_size
    bw = w + border_size
    bh = h + border_size
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    soyaopengl.glColor4f(*color)
    soyaopengl.glVertex2i(x, y)
    soyaopengl.glVertex2i(x, h)
    soyaopengl.glVertex2i(w, h)
    soyaopengl.glVertex2i(w, y)
    soyaopengl.glVertex2i(x, y)
    soyaopengl.glVertex2i(w, y)
    soyaopengl.glColor4f(color[0], color[1], color[2], 0.0)
    soyaopengl.glVertex2i(bw, by)
    soyaopengl.glVertex2i(bx, by)
    soyaopengl.glVertex2i(bx, by)
    soyaopengl.glVertex2i(bx, bh)
    soyaopengl.glColor4f(*color)
    soyaopengl.glVertex2i(x, h)
    soyaopengl.glVertex2i(x, y)
    soyaopengl.glVertex2i(w, h)
    soyaopengl.glVertex2i(x, h)
    soyaopengl.glColor4f(color[0], color[1], color[2], 0.0)
    soyaopengl.glVertex2i(bx, bh)
    soyaopengl.glVertex2i(bw, bh)
    soyaopengl.glVertex2i(bw, bh)
    soyaopengl.glVertex2i(bw, by)
    soyaopengl.glColor4f(*color)
    soyaopengl.glVertex2i(w, y)
    soyaopengl.glVertex2i(w, h)
    soyaopengl.glEnd()

  class MsgWindow(widget.Widget):
    def __init__(self, master = None, title = '', text = '', color = (1.0, 1.0, 1.0, 0.8), font = widget.default_font, bg_color = (0.5, 0.5, 0.5, 0.75)):
      widget.Widget.__init__(self, master)
      self.title = title
      self.text = text
      self.color = color
      self.highlight = (1.0, 1.0, 0.0, 1.0)
      self.font = font
      self.bg_color = bg_color
      self.top = int(soya.get_screen_height() * 0.5)
      self.valid = 0
    def process_event(self, event):
      if (event[0] == soyasdlconst.JOYBUTTONDOWN) or (event[0] == soyasdlconst.MOUSEBUTTONDOWN) or (event[0] == soyasdlconst.KEYDOWN and event[1]):
        if self.master: self.master.children.remove(self)
        go_back()
    def resize(self, p_x, p_y, p_w, p_h):
      self.valid = 0
      self.width = int(soya.get_screen_width() * 0.75)
    def render(self):
      self.left = int((soya.get_screen_width() - self.width) * 0.5)
      if self.valid == 0:
        # HACK
        self.height = self.font.wordwrap(self.title, self.width)[1] + self.font.wordwrap(self.text, self.width)[1]
        soyaopengl.glColor4f(0.0, 0.0, 0.0, 0.0)
        self.font.draw_area(self.title, self.left, 0.0, 0.0, self.width, soya.get_screen_height(), 1)
        self.font.draw_area(self.text , self.left, 0.0, 0.0, self.width, soya.get_screen_height(), 0)
        self.top = int((soya.get_screen_height() - self.height) * 0.5)
        self.valid = 1
      soyaopengl.glEnable(soyaopengl.GL_BLEND)
      soyaopengl.glColor4f(0.0, 0.0, 0.0, 0.5)
      soyaopengl.glBegin(soyaopengl.GL_QUADS)
      soyaopengl.glVertex2i(0, 0)
      soyaopengl.glVertex2i(0, soya.get_screen_height())
      soyaopengl.glVertex2i(soya.get_screen_width(), soya.get_screen_height())
      soyaopengl.glVertex2i(soya.get_screen_width(), 0)
      soyaopengl.glEnd()
      draw_window(self.width, self.height, self.bg_color)
      soyaopengl.glDisable(soyaopengl.GL_BLEND)

      soyaopengl.glColor4f(*self.highlight)
      self.font.draw_area(self.title, self.left, self.top, 0.0, self.width, soya.get_screen_height(), 1)
      self.height = self.font.wordwrap(self.title, self.width)[0]
      
      soyaopengl.glColor4f(*self.color)
      self.font.draw_area(self.text, self.left, self.top + self.height, 0.0, self.width, soya.get_screen_height(), 0)
      self.height += self.font.wordwrap(self.text, self.width)[0]
      
      self.top = int((soya.get_screen_height() - self.height) * 0.5)
      
  class Rectangle(soya.widget.Widget):
    def __init__(self, master = None, color = (0.0, 0.0, 0.0, 0.5)):
      soya.widget.Widget.__init__(self, master)
      self.color = color
      self.visible = 1
    def render(self):
      if self.visible:
        if self.color[3] < 1.0: soyaopengl.glEnable(soyaopengl.GL_BLEND)
        soyaopengl.glColor4f (*self.color)
        soyaopengl.glBegin   (soyaopengl.GL_QUADS)
        soyaopengl.glVertex2f(self.left, self.top)
        soyaopengl.glVertex2f(self.left, self.top + self.height)
        soyaopengl.glVertex2f(self.left + self.width, self.top + self.height)
        soyaopengl.glVertex2f(self.left + self.width, self.top)
        soyaopengl.glEnd()
        if self.color[3] < 1.0: soyaopengl.glDisable(soyaopengl.GL_BLEND)
          
          
  # Inits Soya
  
  #soya.init(width = globdef.SCREEN_WIDTH, height = globdef.SCREEN_HEIGHT, fullscreen = globdef.FULLSCREEN)

  #--------------------#
  # create some shapes #
  #--------------------#

  world = soya.World()
  soya.Face(world, [soya.Vertex(world, -2.0,  0.5, 0.0, 0.0, 0.0),
                    soya.Vertex(world, -2.0, -0.5, 0.0, 0.0, 1.0),
                    soya.Vertex(world,  2.0, -0.5, 0.0, 1.0, 1.0),
                    soya.Vertex(world,  2.0,  0.5, 0.0, 1.0, 0.0)
                    ], soya.Material.get('slune_title'))
  stitle = world.shapify()

  world = soya.World()
  soya.Face(world, [soya.Vertex(world, -0.5,  0.5, 0.0, 0.0, 0.0, (0.0, 0.0, 1.0, 0.75)),
                    soya.Vertex(world, -0.5, -0.5, 0.0, 0.0, 1.0, (0.0, 0.0, 1.0, 0.75)),
                    soya.Vertex(world,  0.5, -0.5, 0.0, 1.0, 1.0, (0.0, 0.0, 1.0, 0.75)),
                    soya.Vertex(world,  0.5,  0.5, 0.0, 1.0, 0.0, (0.0, 0.0, 1.0, 0.75))
                    ], soya.Material.get('halo-noadd'))
  seffect1 = world.shapify()

  world = soya.World()
  soya.Face(world, [soya.Vertex(world, -0.5,  0.5, 0.0, 0.0, 0.0, (1.0, 1.0, 0.0, 1.0)),
                    soya.Vertex(world, -0.5, -0.5, 0.0, 0.0, 1.0, (1.0, 1.0, 0.0, 1.0)),
                    soya.Vertex(world,  0.5, -0.5, 0.0, 1.0, 1.0, (1.0, 1.0, 0.0, 1.0)),
                    soya.Vertex(world,  0.5,  0.5, 0.0, 1.0, 0.0, (1.0, 1.0, 0.0, 1.0))
                     ], soya.Material.get('halo'))
  seffect2 = world.shapify()


  #----------------#
  # create widgets #
  #----------------#

  root = widget.Group()

  #soya.widget.Clearer(root)

  bg_img = soya.widget.Image(root, soya.Material.get('slune_splash'))
  bg_img.resize_style = ('maximize', ('keep ratio', 1024 / 768.0), 'center x', 'center y')
  bg_img.resize_style = ('maximize', 'center x', 'center y')
  bg_rect = Rectangle(root)
  BG_RECT_RESIZE = ('maximize', ('percent top', 0.4), 'maximize height', ('margin bottom', 20), ('margin left', 20), ('margin right', 20))
  bg_rect.resize_style = BG_RECT_RESIZE
  bg_rect.visible = 0

  #--------#
  # banner #
  #--------#

  pub_rect = Rectangle(root)
  pub_rect.resize_style = (('percent left', 0), ('percent top', 0.0), 'maximize width')
  pub_rect.height = soya.widget.default_font.height * 1.3
  pub_banner = soya.widget.HorizontalBanner(root)
  pub_banner.add_elements_from_strings((_('__banner_text__') % globdef.VERSION).split('|'), {}, '  . . .  ')
  
  info_banner = soya.widget.VerticalBanner(root)
  info_banner.add_elements_from_strings(_('__info__').split('|'), {})
  info_banner.speed = 2
  info_banner.resize_style = ('maximize', ('margin left', 20), ('margin right', 20))
  info_banner.loop = 1
  info_banner.visible = 0
  
  
  #-----------------#
  # create 3D scene #
  #-----------------#

  scene = soya.World()
  camera = soya.Camera(scene)
  root.add(camera)
  camera.set_xyz(0.0, -0.5, 5.0)

  atm = soya.NoBackgroundAtmosphere()
  atm.ambient  = (0.2, 0.2, 0.2, 1.0)
  scene.atmosphere = atm
  
  light = soya.Light(scene)
  light.set_xyz(0.0, 1.0, 5.0)
  
  effect1 = soya.Volume(scene, seffect1)
  effect1.set_xyz (-6.0, 0.0, -1.0)
  
  effect2 = soya.Volume(scene, seffect2)
  
  title = soya.Volume(scene, stitle)


  #----------------#
  # menus creation #
  #----------------#

  def set_current_menu(menu):
    global current_menu
    global old_menus
    if current_menu:
      current_menu.visible = 0
      old_menus.append(current_menu)
    current_menu = menu
    current_menu.visible = 1

  def go_back():
    global current_menu
    global old_menus
    if len(old_menus) == 0:
      globdef.generate_dot_slune()
      import sys
      sys.exit()
    if current_menu:
      current_menu.visible = 0
    current_menu = old_menus.pop()
    current_menu.visible = 1

  def popup_msg_window(msg_title, msg_text):
    w = MsgWindow(root, msg_title, msg_text)
    global current_menu
    if current_menu:
      old_menus.append(current_menu)
    current_menu = w
    return w

  def show_h_option():
    menu_h_option.choices[0].value = str(globdef.SCREEN_WIDTH) + 'x' + str(globdef.SCREEN_HEIGHT)
    menu_h_option.choices[1].value = menu_h_option.choices[1].range[globdef.FULLSCREEN]
    menu_h_option.choices[2].value = menu_h_option.choices[2].range[globdef.QUALITY]
    menu_h_option.choices[3].value = globdef.MAX_VISION
    #menu_h_option.choices[4].value = menu_h_option.choices[4].range[globdef.SOUND]
    #menu_h_option.choices[5].value = menu_h_option.choices[5].range[globdef.MUSIC]
    #menu_h_option.choices[6].value = menu_h_option.choices[6].range[globdef.ASYNC_LOAD_MUSIC]
    #menu_h_option.choices[7].value = globdef.SOUND_SYSTEM
    menu_h_option.choices[4].value = str(int(round(globdef.SOUND_VOLUME * 100)))
    set_current_menu(menu_h_option)

  def back_h_option():
    resa = menu_h_option.choices[0].value
    globdef.SCREEN_WIDTH     = int(resa[: resa.find("x")])
    globdef.SCREEN_HEIGHT    = int(resa[resa.find("x") + 1:])
    globdef.FULLSCREEN       = (menu_h_option.choices[1].value == _('on'))
    globdef.QUALITY          = menu_h_option.choices[2].range.index(menu_h_option.choices[2].value)
    globdef.MAX_VISION       = float(menu_h_option.choices[3].value)
    #globdef.SOUND            = (menu_h_option.choices[4].value == _('on'))
    #globdef.MUSIC            = (menu_h_option.choices[5].value == _('on'))
    #globdef.ASYNC_LOAD_MUSIC = (menu_h_option.choices[6].value == _('on'))
    #globdef.SOUND_SYSTEM     = menu_h_option.choices[7].value
    globdef.SOUND_VOLUME     = int(menu_h_option.choices[4].value) / 100.0
    soya.set_quality(globdef.QUALITY)
    soya.set_video(globdef.SCREEN_WIDTH, globdef.SCREEN_HEIGHT, globdef.FULLSCREEN, 1)
    soya.set_sound_volume(globdef.SOUND_VOLUME)
    #import slune.sound; reload(slune.sound)
    go_back()

  def show_p_option():
    menu_p_option.choices[0].value = globdef.NAME
    menu_p_option.choices[1].value = globdef.CHARACTER
    menu_p_option.choices[2].value = menu_p_option.choices[2].range[globdef.VEHICLE]
    menu_p_option.choices[3].value = menu_p_option.choices[3].range[globdef.DIFFICULTY]
    set_current_menu(menu_p_option)

  def back_p_option():
    globdef.NAME       = menu_p_option.choices[0].value
    globdef.CHARACTER  = menu_p_option.choices[1].value
    globdef.VEHICLE    = menu_p_option.choices[2].range.index(menu_p_option.choices[2].value)
    globdef.DIFFICULTY = menu_p_option.choices[3].range.index(menu_p_option.choices[3].value)
    go_back()

  def show_campain():
    set_current_menu(campain)
    
  def show_level():
    menu_level.choices[1].value = globdef.RACE_NB_LAPS
    menu_level.choices[2].value = globdef.RACE_NB_OPPONENTS
    set_current_menu(level)
    
  def show_references():
    pub_banner.visible = 0
    pub_rect.visible = 0
    bg_rect.resize_style = soya.widget.WIDGET_RESIZE_MAXIMIZE
    bg_rect.resize(0, 0, soya.get_screen_width(), soya.get_screen_height())
    info_banner.visible = 1
    info_banner._position = - info_banner.height
    current_menu.visible = 0
    gui_idler.step = -1
    light.visible = 0
    effect1.visible = 0
    effect2.visible = 0
    title.visible = 0
    
  def play_campain():
    import slune.player
    r = slune.player.start_new_game("level-" + select_campain.choices[select_campain.selected])
    if r: # r is the next mission
      globdef.NEXT_MISSION = max(globdef.NEXT_MISSION, int(r[0][8:]))
      missions = map(lambda level_name: int(level_name[14:]), filter(lambda world_name: world_name and world_name.startswith("level-mission"), soya.World.availables()))
      #missions = filter(lambda mission: mission <= globdef.NEXT_MISSION, missions)
      missions.sort()
      missions = map(lambda mission: (_("mission") + "-%s" % mission), missions)
      select_campain.choices = missions
      
      if os.path.exists(os.path.join(soya.path[0], soya.World.DIRNAME, "level-" + r[0] + ".data")):
        select_campain.select(select_campain.selected + 1)
        
    soya.set_root_widget(root)
    
  def play_level():
    #soyaopengl.glEnable(soyaopengl.GL_LIGHTING)
    #soyaopengl.glDisable(soyaopengl.GL_BLEND)
    globdef.RACE_NB_LAPS = menu_level.choices[1].value
    globdef.RACE_NB_OPPONENTS = menu_level.choices[2].value
    import slune.player
    r = slune.player.start_new_game("level-" + select_level.choices[select_level.selected])
    soya.set_root_widget(root)
    
  CHOICELIST_RESIZE_STYLE = ('maximize', ('percent top', 0.4), 'maximize height', ('margin bottom', 20))
  game_menu_1 = soya.widget.ChoiceList(root, [
    soya.widget.Choice(_('Play campain'), show_campain),
    soya.widget.Choice(_('Play level'), show_level),
    soya.widget.Choice(_('Player options'), show_p_option),
    soya.widget.Choice(_('Hardware options'), show_h_option),
    soya.widget.Choice(_('References'), show_references),
    soya.widget.Choice(_('Quit'), go_back)
    ], cancel = 5)
  game_menu_1.resize_style = CHOICELIST_RESIZE_STYLE
  game_menu_1.visible = 0

  checkbox = [_('off'), _('on')]

  menu_h_option = soya.widget.ChoiceList(root, [
    soya.widget.Choice(_('Resolution'), None, None, ['320x240', '640x480', '800x600', '1024x768', '1152x864', '1280x1024', '1400x1050', '1680x1050']),
    soya.widget.Choice(_('Fullscreen'), None, None, checkbox),
    soya.widget.Choice(_('Graphic quality'), None, None, [_('Low'), _('Medium'), _('High')]),
    soya.widget.Choice(_('Max vision distance'), None, None, [0.5, 2.0], 0.1),
    soya.widget.Choice(_('Sound volume'), None, None, ['0', '10', '20', '30', '40', '50', '60', '70', '80', '90', '100']),
    #soya.widget.Choice(_('Sounds'), None, None, checkbox),
    #soya.widget.Choice(_('Music'), None, None, checkbox),
    #soya.widget.Choice(_('Async load music'), None, None, checkbox),
    #soya.widget.Choice(_('Sound system'), None, None, ["OpenAL", "OpenAL (old)", "SDL_mixer"]),
    soya.widget.Choice(_('Back'), back_h_option)
    ], cancel = -1)
  menu_h_option.resize_style = CHOICELIST_RESIZE_STYLE
  menu_h_option.visible = 0

  menu_p_option = soya.widget.ChoiceList(root, [
    soya.widget.ChoiceInput(_('Name')),
    soya.widget.Choice(_('Character'), None, None, map(_, globdef.CHARACTERS)),
    soya.widget.Choice(_('Vehicle'), None, None, [_('Biplane'), _('Car'), _('Truck'), _('Scooter'), _('Tanker')]),
    soya.widget.Choice(_('Difficulty'), None, None, [_('Newbie'), _('Hacker'), _('Guru')]),
    soya.widget.Choice(_('Back'), back_p_option)
    ], cancel = 4)
  menu_p_option.resize_style = CHOICELIST_RESIZE_STYLE
  menu_p_option.visible = 0

  campain = widget.Group(root)
  campain.visible = 0
  def campain_process_event(event):
    if (event[0] == soyasdlconst.MOUSEMOTION):
      if (event[1] < campain.width * 0.5):
        select_campain.enabled = 1
        menu_campain.selected = -1
      else:
        select_campain.enabled = 0
    elif (event[0] == soyasdlconst.KEYDOWN):
      if   (event[1] == soyasdlconst.K_LEFT):
        select_campain.enabled = 1
        menu_campain.selected = -1
        return
      elif (event[1] == soyasdlconst.K_RIGHT):
        select_campain.enabled = 0
        menu_campain.selected = 0
        return
      elif (event[1] == soyasdlconst.K_ESCAPE):
        go_back()
        return
    elif (event[0] == soyasdlconst.JOYAXISMOTION) and (event[1] == 0):
      if   event[2] < 0:
        select_campain.enabled = 1
        menu_campain.selected = -1
        return
      elif event[2] > 0:
        select_campain.enabled = 0
        menu_campain.selected = 0
        return
      
    if (select_campain.enabled == 1):
      select_campain.process_event(event)
    else:
      menu_campain.process_event(event)
  campain.process_event = campain_process_event
  del campain_process_event
  
  missions = map(lambda level_name: int(level_name[14:]), filter(lambda world_name: world_name and world_name.startswith("level-mission"), soya.World.availables()))
  #missions = filter(lambda mission: mission <= globdef.NEXT_MISSION, missions)
  missions.sort()
  missions = map(lambda mission: (_("mission") + "-%s" % mission), missions)
  select_campain = soya.widget.Selector(campain, missions, _('Select level'), arrows = soya.Material.get('widget_arrow'))
  SELECTOR_RESIZE_STYLE = (('percent top', 0.4), 'maximize height', ('margin bottom', 20), ('percent left', 0.0), ('percent width', 0.5))
  select_campain.resize_style = SELECTOR_RESIZE_STYLE
  
  HALF_CHOICELIST_RESIZE_STYLE = (('percent left', 0.5), 'maximize width', ('percent top', 0.4), 'maximize height', ('margin bottom', 20))

  menu_campain = soya.widget.ChoiceList(campain, [
    soya.widget.Choice(_('Back'), go_back),
    soya.widget.Choice(_('Start'), play_campain),
    ], cancel = 0)
  menu_campain.resize_style = HALF_CHOICELIST_RESIZE_STYLE

  level = widget.Group(root)
  level.visible = 0
  def level_process_event(event):
    if (event[0] == soyasdlconst.MOUSEMOTION):
      if (event[1] < level.width * 0.5):
        select_level.enabled = 1
        menu_level.selected = -1
      else:
        select_level.enabled = 0
    elif (event[0] == soyasdlconst.KEYDOWN and (select_level.enabled == 1 or menu_level.selected == 0 or menu_level.selected == 3)):
      if (event[1] == soyasdlconst.K_LEFT):
        select_level.enabled = 1
        menu_level.selected = -1
        return
      elif (event[1] == soyasdlconst.K_RIGHT):
        select_level.enabled = 0
        menu_level.selected = 0
        return
      elif (event[1] == soyasdlconst.K_ESCAPE):
        go_back()
        return
    elif (event[0] == soyasdlconst.JOYAXISMOTION) and (event[1] == 0):
      if   event[2] < 0:
        select_level.enabled = 1
        menu_level.selected = -1
        return
      elif event[2] > 0:
        select_level.enabled = 0
        menu_level.selected = 0
        return

    if (select_level.enabled == 1):
      select_level.process_event(event)
    else:
      menu_level.process_event(event)
  level.process_event = level_process_event
  del level_process_event

  levels = map(lambda level_name: level_name[6:], filter(lambda world_name: world_name and world_name.startswith("level-") and not world_name.startswith("level-mission"), soya.World.availables()))
  select_level = soya.widget.Selector(level, levels, _('Select level'), arrows = soya.Material.get('widget_arrow'))
  select_level.resize_style = SELECTOR_RESIZE_STYLE

  menu_level = soya.widget.ChoiceList(level, [
    soya.widget.Choice(_('Back'), go_back),
    soya.widget.Choice(_('Lap number'), None, None, [1, 8], 1),
    soya.widget.Choice(_('Opponents'), None, None, [2, 6], 1),
    soya.widget.Choice(_('Start'), play_level),
    ], cancel = 0)
  menu_level.resize_style = HALF_CHOICELIST_RESIZE_STYLE


  #----------#
  # start !! #
  #----------#
  
  soya.set_root_widget(root)
  
  class Idler(soya.Idler):
    ltime = 0.0
    step = 0
    nb_step = 2
    
    def begin_round(self):
      soya.Idler.begin_round(self)
      if self.step == -1:
        # reference
        for event in soya.process_event():
          if (event[0] == soyasdlconst.KEYDOWN and event[1]) or (event[0] == soyasdlconst.MOUSEBUTTONDOWN) or (event[0] == soyasdlconst.JOYBUTTONDOWN):
            info_banner.visible = 0
            pub_banner.visible = 1
            pub_rect.visible = 1
            bg_rect.resize_style = BG_RECT_RESIZE
            bg_rect.resize(0, 0, soya.get_screen_width(), soya.get_screen_height())
            current_menu.visible = 1
            self.step = 3
            light.visible = 1
            effect1.visible = 1
            effect2.visible = 1
            title.visible = 1
            break
      elif self.step < 2:
        for event in soya.process_event():
          if (event[0] == soyasdlconst.JOYBUTTONDOWN) or (event[0] == soyasdlconst.KEYDOWN and event[1]) or (event[0] == soyasdlconst.MOUSEBUTTONDOWN):
            self.end_step_1()
            self.end_step_2()
            self.step = 3
            break
      elif current_menu:
        for event in soya.process_event(): current_menu.process_event(event)
          
    def end_step_1(self):
      effect1.set_identity()
      effect1.scale(20.0, 4.0, 4.0)
      effect1.rotate_incline(180.0)
      effect1.set_xyz(0.0, 0.0, -1.0)
      effect2.set_identity()
      effect2.scale(4.0, 0.1, 0.1)
      effect2.set_xyz(0.0, -0.6, -1.0)
      title.set_identity()

    def end_step_2(self):
      camera.set_xyz(0.0, -1.5, 5.0)
      set_current_menu(game_menu_1)
      bg_rect.visible = 1

    def advance_time(self, proportion):
      if self.step == -1:
        info_banner.advance_time(proportion)
      else:
        pub_banner.advance_time(proportion)

      if self.step == 0:
        if self.ltime < 150.0:
          self.ltime = self.ltime + proportion
          percent = self.ltime / 150.0

          effect1.set_identity()
          effect1.scale(20.0 * percent, 4.0 * percent, 4.0 * percent)
          effect1.rotate_incline(180.0 * percent)
          effect1.set_xyz(0.0, 0.0, -1.0)

          effect2.set_identity()
          effect2.scale(4.0 * percent, 0.1 * percent, 0.1 * percent)
          effect2.set_xyz(0.0, -0.6, -1.0)

          title.set_identity()
          title.scale(percent, percent, percent)
          title.rotate_incline(360.0 * percent)
        else:
          self.end_step_1()
          self.step = 1
          self.ltime = 0.0

      elif self.step == 1:
        if self.ltime < 40.0:
          self.ltime = self.ltime + proportion
          percent = self.ltime / 40.0
          camera.set_xyz(0.0, -0.5 - 1.0 * percent, 5.0)
        else:
          self.end_step_2()
          self.step = 2
          self.ltime = 0.0

    def render(self):
      soya.render()


  gui_idler = Idler()
  action = gui_idler.idle()
  
  
  



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This script is used for genetically optimizing racing IA.
# It considers 2 character at a time; see race_algogen for a version with a single IA.

from __future__ import generators

RENDER = 0
PERSO  = "tux"
LEVEL  = "crashfunkel-race-2"
import sys

i = 1
while i < len(sys.argv):
  arg = sys.argv[i]
  if   arg == "--render": RENDER = 1
  
  elif arg == "--perso":
    i += 1
    PERSO = sys.argv[i]
  
  elif arg == "--level":
    i += 1
    LEVEL = sys.argv[i]
  
  elif arg == "--help":
    print """Optimize racing IA, by running 2 IAs against each other.
The resulting chromosoms are printed on stdout, and should be manually copied in race.py.

Options:
  --render         Render the 3D, so as you can see something (else no rendering is performed, for speedup).
  --perso <perso>  Choose the character IA to optimize (e.g. tux, gnu,...).
  --level <level>  Choose the level. Should be a racing level."""
    sys.exit()
    
  else:
    print "Unknown options %s !" % arg
    sys.exit(1)
    
  i += 1


import random
import soya, soya.soya3d as soya3d, soya.game.level as soya_game_level, soya.game.camera as soya_camera
import py2play.level
import slune.race as race, slune.level, slune.character, slune.player

MAX_DURATION = 1500

class Idler(slune.level.Idler):
  def __init__(self):
    slune.level.Idler.__init__(self)
    self.next_flag_name1 = ""
    self.next_flag_name2 = ""
    
  def idle(self):
    value1 = value2 = MAX_DURATION
    OPPONENT1.best_lap = OPPONENT2.best_lap = 9999.0
    race.init_race_opponent(OPPONENT1)
    race.init_race_opponent(OPPONENT2)
    
    for nb_round in range(MAX_DURATION):
      self.begin_round()
      self.advance_time(1.0)
      self.end_round()
      
      if self.next_round_tasks:
        for task in self.next_round_tasks: task()
        self.next_round_tasks = []
        
      self.render()

      if OPPONENT1.best_lap != 9999.0:
        if value1 == MAX_DURATION: value1 = nb_round
      else:
        if OPPONENT1.life < 0: value1 = MAX_DURATION + 1
        
      if OPPONENT2.best_lap != 9999.0:
        if value2 == MAX_DURATION: value2 = nb_round
      else:
        if OPPONENT2.life < 0: value2 = MAX_DURATION + 1
        
      if (value1 != MAX_DURATION) and (value2 != MAX_DURATION): break
      
    return value1, value2
  
  def render(self):
    if RENDER: slune.level.Idler.render(self)
    
    if hasattr(OPPONENT1, "next_flag_name"):
      if self.next_flag_name1 != OPPONENT1.next_flag_name:
        self.next_flag_name1 = OPPONENT1.next_flag_name
        print "Opponent 1: Next flag: ", self.next_flag_name1
        
    if hasattr(OPPONENT2, "next_flag_name"):
      if self.next_flag_name2 != OPPONENT2.next_flag_name:
        self.next_flag_name2 = OPPONENT2.next_flag_name
        print "Opponent 2: Next flag: ", self.next_flag_name2
        
  
soya.init(width = 640, height = 480, fullscreen = 0)

idler = Idler()

import py2play.player
py2play.player.CURRENT_PLAYER = 1 # HACK !

level = py2play.level.CREATE("level-" + LEVEL)
level.set_active(1)
level.max_laps = 999
for character in level.characters: level.remove_character(character)

class Opponent(race.Opponent):
  def __init__(self, *args, **kargs):
    race.Opponent.__init__(self, *args, **kargs)
    
    self.value = 0
    
OPPONENT1 = None
OPPONENT2 = None
def test(opponent1, opponent2):
  global OPPONENT1, OPPONENT2
  OPPONENT1 = opponent1
  OPPONENT2 = opponent2
  
  OPPONENT1.perso_name = PERSO
  OPPONENT2.perso_name = PERSO
  
  idler.camera.add_traveling(soya_camera.ThirdPersonTraveling(OPPONENT2.camera_target))
  
  OPPONENT1.set_level(level, 1)
  OPPONENT2.set_level(level, 1)
  
  value1, value2 = idler.idle()

  # If they are not dead !
  if OPPONENT1.parent: OPPONENT1.parent.remove_character(OPPONENT1)
  if OPPONENT2.parent: OPPONENT2.parent.remove_character(OPPONENT2)
  
  OPPONENT1.level = None
  OPPONENT2.level = None
  
  return value1, value2

def value_sorter(a, b): return cmp(a.value, b.value)

def mutate(value):
  if random.random < 0.6: return value
  if not value: return 0.02
  return value + (random.random() - 0.5) * value

def make_love(father, mother):
  chromosom = map(random.choice, zip(father.chromosom, mother.chromosom))
  chromosom = map(mutate, chromosom)
  return Opponent(chromosom = chromosom)



OPPONENTS = [Opponent(chromosom = race.CHROMOSOMS[(PERSO, 2)]), Opponent(chromosom = race.CHROMOSOMS[(PERSO, 2)])]

while 1:
  for i in range(4):
    OPPONENTS.append(make_love(random.choice(OPPONENTS), random.choice(OPPONENTS)))
    
  for i in range(0, len(OPPONENTS), 2):
    opponent1 = OPPONENTS[i]
    opponent2 = OPPONENTS[i + 1]
    opponent1.value, opponent2.value = test(opponent1, opponent2)
    
  OPPONENTS.sort(value_sorter)
  OPPONENTS = OPPONENTS[:4]
  
  for opponent in OPPONENTS:
    print opponent.chromosom, "=>", opponent.value



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import os.path, time, random, cPickle as pickle
import py2play.level as level
import py2play.idler as idler
import soya, soya.widget as widget, soya.opengl as soyaopengl
import slune.globdef as globdef, slune.character, slune.sound as sound

from soya import Point, Vector

# class TravelingCamera(soya.TravelingCamera):
#   def begin_round(self):
#     pass

#   def advance_time(self, proportion):
#     pass

class FPSLabel(widget.Widget):
  """FPSLabel

A label that shows the FPS."""
  
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.x = parent_left + parent_width  - 150
    self.y = parent_top  + parent_height - 37
    
  def render(self):
    soya.DEFAULT_MATERIAL.activate()
    soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    if idler.IDLER:
      widget.default_font.draw("%.1f FPS" % idler.IDLER.fps, self.x, self.y)
    else:
      widget.default_font.draw("No idler", self.x, self.y)
      
      
class BlackBands(widget.Widget):
  """Black bands for 16/9 look."""
  
  def render(self):
    w = float(soya.get_screen_width ())
    h = float(soya.get_screen_height())
    t = h / 10.0
    soyaopengl.glColor4f(0.0, 0.0, 0.0, 1.0)
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    
    soyaopengl.glVertex2f(0.0, 0.0)
    soyaopengl.glVertex2f(0.0, t)
    soyaopengl.glVertex2f(w, t)
    soyaopengl.glVertex2f(w, 0.0)
    
    soyaopengl.glVertex2f(0.0, h - t)
    soyaopengl.glVertex2f(0.0, h)
    soyaopengl.glVertex2f(w, h)
    soyaopengl.glVertex2f(w, h - t)
    
    soyaopengl.glEnd()
    
#BLACK_BANDS =  BlackBands()


class Idler(soya.World, idler.Idler):
  #def __del__(self):
  #  print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! del Idler"
  will_render = 1
  
  def __init__(self, camera = None):
    soya .World.__init__(self)
    idler.Idler.__init__(self)
    
    self.camera = camera or soya.TravelingCamera(self)
    soya.set_root_widget(widget.Group())
    soya.root_widget.add(self.camera)

    #def f():
    #  print "!"
    #self.camera.begin_round = f
    
    sound.init(self.camera)
    
    self.blackbands_group    = widget.Group() 
    self.no_blackbands_group = widget.Group()
    soya.root_widget.add(self.no_blackbands_group)
    
    self.lifebar = slune.character.LifeBar()
    self.no_blackbands_group.add(self.lifebar)
    
    self.blackbands_group.add(BlackBands())
    
    self.blackbands_visible = 0
    
    self.players_info = []
    self.waiting_until_ready = 0
    self.paused = None
    
    soya.IDLER     = self
    soya.MAIN_LOOP = self
    
    text = soya.widget.Label(None, "", 1)
    
    def text_resize(parent_left, parent_top, parent_width, parent_height):
      text.width  = parent_width - 50 # The lifebar takes about 50 pixels !
      text.height = int(parent_height * 0.5)
      text.left   = parent_left
      text.top    = parent_top
      
    text.resize = text_resize
    soya.root_widget.add(text)
    self.messageboard = text
    self.messages = []
    
    if globdef.VERIFICATION_SERVER: self.logon_verif()
    
  def show_fps(self): soya.root_widget.add(FPSLabel())
  
  def show_blackbands(self):
    if not self.blackbands_visible:
      self.blackbands_visible = 1
      h = soya.get_screen_height()
      h2 = h / 10
      soya.root_widget.remove(self.no_blackbands_group)
      soya.root_widget.resize(0, h2, soya.get_screen_width(), h - 2 * h2)
      soya.root_widget.insert(1, self.blackbands_group)
      
  def hide_blackbands(self):
    if self.blackbands_visible:
      self.blackbands_visible = 0
      soya.root_widget.remove(self.blackbands_group)
      soya.root_widget.resize(0, 0, soya.get_screen_width(), soya.get_screen_height())
      soya.root_widget.add   (self.no_blackbands_group)
      #self.messageboard.font._image().save("/home/jiba/tmp/font.png")
    
    
  def logon_verif(self):
    """Log on the verif server."""
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((globdef.VERIFICATION_SERVER, 36081))
    self.verif = s.makefile("w", 0)
    self.verif.write("R") # Check rounds value
    
  def idle(self):
    self.message(_("__ready__"))
    
    from slune.controler import KeyboardMouseControler
    
    self.waiting_until_ready = 1
    wait = WaitUntilReady(self)
    
    for level in self:
      if isinstance(level, Level):
        for character in level.characters:
          if character.controler:
            character.controler.animate(wait)
            if isinstance(character.controler, KeyboardMouseControler):
              character.controler.enabled = 0
              
    return idler.Idler.idle(self)
    
  def pause(self, max_duration = -1): self.next_round_tasks.append(lambda : self._pause(max_duration))
  
  def _pause(self, max_duration = -1):
    if not self.paused:
      from slune.controler import KeyboardMouseControler

      self.paused = wait = WaitClick(self, max_duration)

      for level in self:
        if isinstance(level, Level):
          for character in level.characters:
            if character.controler:
              character.controler.animate(wait)
              if isinstance(character.controler, KeyboardMouseControler):
                character.controler.enabled = 0
              
  def unpause(self): self.next_round_tasks.append(self._unpause)
  
  def _unpause(self):
    if self.paused:
      from slune.controler import KeyboardMouseControler
      
      for level in self:
        if isinstance(level, Level):
          for character in level.characters:
            if isinstance(character.controler, KeyboardMouseControler):
              character.controler.enabled = 1
              
      self.paused.stop() # Stop the WaitClick instance !
      self.paused = None
      
      
  def init_game(self):
    self.camera.zap()
    
    for o in self.recursive():
      if isinstance(o, soya.World) and o.atmosphere:
        if o.atmosphere.fog:
          self.camera.back = o.atmosphere.fog_end = o.atmosphere.fog_end * globdef.MAX_VISION
          break
        
  def start_game(self):
    if self.waiting_until_ready:
      self.waiting_until_ready = 0
      self.clear_message()
      
      from slune.controler import KeyboardMouseControler
      
      for level in self:
        if isinstance(level, Level):
          for character in level.characters:
            if isinstance(character.controler, KeyboardMouseControler):
              character.controler.controlers[-1].stop() # Stop the WaitUntilReady instance !
              character.controler.enabled = 1
              
      from py2play.player import CURRENT_PLAYER
      CURRENT_PLAYER.level.start_game()
      
  def level_completed(self, character, winner = None, message = None):
    import videosequence
    
    videosequence.MissionCompleted(winner, message).start(tux = character)
    
  def begin_round(self):
    time.sleep(0.00001) # for sound
    
    # In THIS order !!! Say to Py2Play the round begin before doing anything !
    idler.Idler.begin_round(self)
    
    # The camera must be played AFTER the character, for a better visual effect (ie to avoid the character to move after the camera has been moved)
    for child in self.children: child.begin_round()
      
    soya.root_widget.widget_begin_round()
    
    if globdef.VERIFICATION_SERVER:
      import py2play.player, cPickle as pickle
      
      players_rounds = {}
      for level in self:
        if isinstance(level, Level):
          for character in level.characters:
            if hasattr(character, "player"):
              players_rounds[character.player.name] = character.round
              
      pickle.dump((players_rounds, py2play.player.CURRENT_PLAYER.name), self.verif)
      
  def advance_time(self, proportion):
    #soya.advance_time(proportion)
    for child in self.children: child.advance_time(proportion)
    soya.root_widget.widget_advance_time(proportion)
    
  def render(self):
    for i in soya.BEFORE_RENDER: i()
    soya.render()
    
  def end_round(self):
    current = time.time()
    if self.messages and (self.messages[0][1] <= current):
      del self.messages[0]
      self.messageboard.text = "\n".join([text for (text, duration) in self.messages])
      
    # In THIS order !!! Really ends the round before saying to Py2Play it is ended !
    # The camera must be played AFTER the character, for a better visual effect (ie to avoid the character to move after the camera has been moved)
    for child in self.children:
      if hasattr(child, "end_round"): child.end_round()
      
    soya.root_widget.widget_end_round()
    
    idler.Idler .end_round(self)
    
  def clear_message(self):
    self.messages *= 0
    self.messageboard.text = ""
    
  def message(self, text, duration = 7.0):
    self.messages.append((text, duration + time.time()))
    
    if len(self.messages) > 3: del self.messages[0]
    
    self.messageboard.text = "\n".join([text for (text, duration) in self.messages])
    
  def message_append(self, text, duration = 7):
    self.messages[-1] = (self.messages[-1][0] + text, duration + time.time())
    
    self.messageboard.text = "\n".join([text for (text, duration) in self.messages])
    
class WaitUntilReady:
  def __init__(self, idler):
    self._stop = 0
    self.idler = idler
    
    self.players_info = ()
    
  def stop(self): self._stop = 1
  
  def next(self):
    if self._stop: raise StopIteration
    
    from py2play.player import CURRENT_PLAYER
    
    if self.players_info != CURRENT_PLAYER.players_info:
      for player_info in CURRENT_PLAYER.players_info:
        if not player_info in self.players_info:
          self.idler.message(_("__connection__") % (player_info[0], player_info[1]))
      self.players_info = CURRENT_PLAYER.players_info[:]
        
    return slune.character.Action()
  
class WaitClick:
  def __init__(self, idler, duration = -1):
    self._stop = 0
    self.idler = idler
    self.duration = duration
  def stop(self): self._stop = 1
  
  def next(self):
    if self._stop: raise StopIteration
    if not self.duration:
      from py2play.idler import IDLER
      IDLER.unpause()
    self.duration -= 1
    return slune.character.Action()



class MessageBoard(soya.widget.Label):
  def __init__(self, master = None):
    soya.widget.Label.__init__(self, master, "", 1)
    self.messages = []
    self.left   = 0
    self.top    = 0
    self.width  = 200
    self.height = 200
  def add_message(self, text, color = soya.WHITE, time = 300):
    self.messages.append(text)
    
    if len(self.messages) > 3:
      del self.messages[0]
      
    self.text = "\n".join(self.messages)
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.left   = 0
    self.top    = 0
    self.width  = parent_width
    self.height = int(parent_height * 0.5)
    
    
class Level(level.Level, soya.World):
  pushables = () # HACK ! No push before the game really starts !
  bonuss    = () # HACK !
  
  #def __del__(self):
  #  print "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! del Level", self.name
  
  def set_active(self, active):
    level.Level.set_active(self, active)
    
    # The camera must be played AFTER the character, for a better visual effect (ie to avoid the character to move after the camera has been moved)
    # => insert the level at the position 0 (=before the camera)
    if self.active: idler.IDLER.insert(0, self)
    else:           idler.IDLER.remove(self)
    
  def __init__(self, parent = None):
    level.Level.__init__(self, "")
    soya .World.__init__(self, parent)
    
    self.init_script = self.init_character_script = self.start_script = self.end_script = ""
    
  def init_character(self, character):
    if isinstance(character, slune.character.Competitor) and character.controler:
      character.rank = len([character for character in self.characters if isinstance(character, slune.character.Competitor)]) - 1
      
      exec self.init_character_script in globals(), locals()
      
      character.y += character.floating
      
    character.init()
    
  def remove_character(self, character):
    if self.pushables and isinstance(character, slune.character.Pusher):
      try: self.pushables.remove(character)
      except: pass
      
    level.Level.remove_character(self, character)
    self.remove(character)
    
  def __setstate__(self, state):
    level.Level.__setstate__(self, state)
    
    from py2play.player import CURRENT_PLAYER
    if CURRENT_PLAYER:
      if not hasattr(self, "inited"): # Else already inited, e.g. by another player (in network game)
        self.inited     = 1
        self.random     = random.Random()
        self.difficulty = globdef.DIFFICULTY # Store difficulty in level, because, in network gaming, all players plays at the same difficulty !
        
        exec self.init_script
        
      if hasattr(self, "music_name"):
        if self.music_name: sound.play_music(self.music_name)
      else:
        sound.play_music("summum_of_the_light.ogg") # Default music
        
  def start_game(self):
    exec self.start_script in globals(), locals()
    
    self.pushables = [x for x in self if isinstance(x, slune.character.Pusher)]
    self.bonuss    = [x for x in self if isinstance(x, Bonus                 )]
    
    for pushable in [x for x in self if isinstance(x, slune.character.Pushable)]:
      h = self.random.random()
      if   h < pushable.bonus_chance:
        pushable.bonus_type = 1
      elif h < 2 * pushable.bonus_chance:
        pushable.bonus_type = 2
        
  def end_game(self):
    exec self.end_script in globals(), locals()
    
    sound.end_music()
    
  def begin_round(self):
    for child in self.children:
      
      #if child.__class__.__name__ in ["Bowling", "World", "Pushable", "PlayerCharacter"]:
      #if child.__class__.__name__ in ["Bowling", "World", "Pushable"]:
      #if child.__class__.__name__ in ["PlayerCharacter"]:
        child.begin_round()
        
        
def CREATE(level_name):
  if isinstance(level_name, unicode): level_name = level_name.encode("latin")
  lev = pickle.load(open(os.path.join(soya.path[0], soya.World.DIRNAME, level_name + ".data"), "rb"))
  lev.name = level_name
  level.store(lev)
  
  return lev
level.CREATE = CREATE


class SphericalScript(soya.Volume):
  def __init__(self, parent = None, script = ""):
    soya.Volume.__init__(self, parent, soya.Shape.get("sphericalscript"))
    self.script   = script
    self.radius   = 1.0
    self.disabled = 0
    
  def set_radius(self, radius):
    r = radius / self.radius
    self.scale(r, r, r)
    self.radius = radius
    
  def disable(self, nb_round = None):
    if nb_round is None: self.parent.remove(self)
    else: self.disabled += nb_round
    
  def __setstate__(self, state):
    soya.Volume.__setstate__(self, state)
    
    from py2play.player import CURRENT_PLAYER
    if CURRENT_PLAYER: self.set_shape(None)
    
  def begin_round(self):
    if not self.shape:
      if self.disabled: self.disabled -= 1
      else:
        self.set_shape(None)
        for character in self.parent.characters:
          if character.distance_to(self) <= self.radius:
            exec self.script in globals(), locals()


class Bonus:
  come_back = 1

class LifeBonus(Bonus, soya.Bonus):
  radius = 3.0
  
  def __init__(self, parent = None):
    soya.Bonus.__init__(self, parent, soya.Material.get("life1"), soya.Material.get("halo"))
    self.color    = (1.0, 0.40000000596046448, 0.20000000298023224, 1.0)
    self.disabled = 0
    
  def begin_round(self):
    if self.disabled:
      self.disabled -= 1
      if not self.disabled: self.visible = 1
      
    else:
      for character in self.parent.characters:
        if self.distance_to(character) < self.radius:
          character.life = 1.0
          sound.play("bonus-1.wav", character)
          
          LifeBonusParticle(character.parent).set_xyz(self.x, self.y + 0.3, self.z)
          
          if self.come_back:
            self.disabled  = 150
            self.visible   = 0
          else:
            self.parent.bonuss.remove(self)
            self.parent.remove(self)
          break
        
class FlameThrowerBonus(Bonus, soya.Bonus):
  radius = 3.0
  
  def __init__(self, parent = None):
    soya.Bonus.__init__(self, parent, soya.Material.get("bonus2"), soya.Material.get("halo"))
    self.color    = (0.5, 1.0, 0.5, 1.0)
    self.disabled = 0
    
  def begin_round(self):
    if self.disabled:
      self.disabled -= 1
      if not self.disabled: self.visible = 1
      
    else:
      for character in self.parent.characters:
        if self.distance_to(character) < self.radius:
          character.flame_thrower += 2
          sound.play("bonus-1.wav", character)
          
          LifeBonusParticle(character.parent).set_xyz(self.x, self.y + 0.3, self.z)
          
          if self.come_back:
            self.disabled  = 150
            self.visible   = 0
          else:
            self.parent.bonuss.remove(self)
            self.parent.remove(self)
          break
        
import math, soya

class LifeBonusParticle(soya.Particles):
  def __init__(self, parent = None, material = None, nb_particles = 30):
    soya.Particles.__init__(self, parent, material, nb_particles, 1)
    self.set_colors((0.2, 0.0, 0.1, 1.0), (1.0, 0.2, 0.6, 1.0), (0.1, 0.0, 0.0, 1.0))
    self.set_sizes ((1.0, 1.0))
    self.life = 10
    #self.regenerate()
    
  def generate(self, index):
    angle = random.random() * 3.1417
    sx =  math.cos (angle)
    sy = -math.sqrt(random.random() * 3.0)
    sz =  math.sin (angle)
    l = (0.06 * (1.5 + random.random())) / math.sqrt(sx * sx + sy * sy + sz * sz)
    self.set_particle(index, 0.5 + random.random(), sx * l, sy * l, sz * l, 0.0, 0.03, 0.0)
    
  def aaabegin_round(self):
    soya.Particles.begin_round(self)
    if   self.life >  0: self.life -= 1
    elif self.life == 0:
      self.auto_generate_particle = 0
      #self.removable = 1
      self.life = -1




# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya, soya.widget as widget, soya.opengl as soyaopengl
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
import slune.character as slune_character
import slune.controler
import slune.globdef as globdef, slune.sound as sound, slune.level as slune_level

class Medoc(soya.Volume):
  radius = 2.5
  
  def __init__(self, parent = None):
    soya.Volume.__init__(self, parent, soya.Shape.get("medoc2"))
    
  def begin_round(self):
    for character in self.parent.characters:
      if isinstance(character, slune_character.Competitor) and (character.distance_to(self) < self.radius):
        sound.play("flag-1.wav", character)
        firework = soya.Smoke(character.level, nb_particles = 10, removable = 1)
        firework.move(self)
        firework.regenerate()
        
        if hasattr(character, "nb_medoc"):
          character.nb_medoc += 1
          if character.nb_medoc >= self.nb_max:
            from py2play.idler import IDLER
            IDLER.level_completed(character, 1, _("__medoc_winner__"))
            
        else:
          character.nb_medoc = 1
          #if character.vehicle == 2:
          #  character.medoc = soya.Volume(character.internal, soya.Shape.get("medoc2"))
            
        self.parent.remove(self)
        

def add_medocs(level, nb, minimap = 1):
  down = soya.Vector(level, 0.0, -1.0, 0.0)
  
  for i in range(nb):
    m = Medoc(level)
    m.nb_max = nb
    while 1:
      m.set_xyz(level.random.uniform(5.0, 200.0), 1000.0, level.random.uniform(5.0, 200.0))
      
      r = level.raypick(m, down, -1, 3)
      if r:
        r[0].convert_to(level)
        m.y = r[0].y + 0.5
        break
      
  from py2play.idler import IDLER
  if minimap: IDLER.no_blackbands_group.insert(0, MiniMap(level))
  

class MiniMap(widget.Widget):
  def __init__(self, level):
    self.level = level
    self.point = soya.Point()
    
    self.top = 10
    self.width  = 200
    self.height = 200
    self.left = soya.root_widget.width - self.width - 40
    
    medocs = [x for x in level if isinstance(x, Medoc)]
    min_x = min_y =  10e1000
    max_x = max_y = -10e1000
    for medoc in medocs:
      if medoc.x < min_x: min_x = medoc.x
      if medoc.x > max_x: max_x = medoc.x
      if medoc.z < min_y: min_y = medoc.z
      if medoc.z > max_y: max_y = medoc.z
    f_x = self.width  / (max_x - min_x)
    f_y = self.height / (max_y - min_y)
    if f_x < f_y:
      f = f_x
      left = self.left
    else:
      f = f_y
      left = self.left + ((max_y - min_y) - (max_x - min_x)) * f
    top = self.top
    
    self.f     = f
    self.cleft = left
    self.ctop  = top
    self.min_x = min_x
    self.min_y = min_y
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.left = parent_width - self.width - 40
    self.screen_width  = parent_width
    self.screen_height = parent_height
    
  def render(self):
    soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    for medoc in [x for x in self.level if isinstance(x, Medoc)]:
      x = self.cleft + self.f * (medoc.x - self.min_x)
      y = self.ctop  + self.f * (medoc.z - self.min_y)
      soyaopengl.glVertex2f(x - 4.0, y - 4.0)
      soyaopengl.glVertex2f(x - 4.0, y + 4.0)
      soyaopengl.glVertex2f(x + 4.0, y + 4.0)
      soyaopengl.glVertex2f(x + 4.0, y - 4.0)
      
    soyaopengl.glEnd()
    
    for character in self.level.characters:
      if isinstance(character, slune_character.Competitor):
        x  = self.cleft + self.f * (character.x - self.min_x)
        y  = self.ctop  + self.f * (character.z - self.min_y)
        
        self.point.__init__(character, 0.0, 0.0, -25.0)
        self.point.convert_to(self.level)
        x2 = self.cleft + self.f * (self.point.x - self.min_x)
        y2 = self.ctop  + self.f * (self.point.z - self.min_y)
        
        soya.DEFAULT_MATERIAL.activate()
        soyaopengl.glBegin(soyaopengl.GL_LINES)
        soyaopengl.glVertex2f(x , y )
        soyaopengl.glVertex2f(x2, y2)
        soyaopengl.glEnd()
        
        material = soya.Material.get("head_" + character.perso_name + "_1")
        material.activate()
        soyaopengl.glEnable(soyaopengl.GL_BLEND)
        soyaopengl.glBegin(soyaopengl.GL_QUADS)
        soyaopengl.glTexCoord2f(0.0, 0.0); soyaopengl.glVertex2f(x - 20.0, y - 20.0)
        soyaopengl.glTexCoord2f(0.0, 1.0); soyaopengl.glVertex2f(x - 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 1.0); soyaopengl.glVertex2f(x + 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 0.0); soyaopengl.glVertex2f(x + 20.0, y - 20.0)
        soyaopengl.glEnd()
        soyaopengl.glDisable(soyaopengl.GL_BLEND)
        soya.DEFAULT_MATERIAL.activate()



class MedocWaiter(soya.Volume):
  def __init__(self, parent, x, y, z):
    soya.Volume.__init__(self, parent, soya.Shape.get("girafe1"))
    self.set_xyz(x, y, z)
    self.medoc_received = 0
    
  def begin_round(self):
    if not self.medoc_received:
      from py2play.player import CURRENT_PLAYER
      
      if CURRENT_PLAYER.character.distance_to(self) < 10.0:
        self.medoc_received = 1
        sound.play("flag-1.wav", self)
        GiveMedoc(self.parent, CURRENT_PLAYER.character, self)
        
        
class GiveMedoc(soya.Volume):
  def __init__(self, parent, character, receiver):
    soya.Volume.__init__(self, parent, soya.Shape.get("medoc2"))
    self.move   (soya.Point (character, 0.0, 0.5,  0.3))
    self.look_at(soya.Vector(character, 0.0, 0.0, -1.0))
    self.receiver = receiver
    
  def begin_round(self):
    if self.distance_to(self.receiver) < 1.0:
      for e in self.parent:
        if isinstance(e, MedocWaiter) and (e.medoc_received == 0): break
      else:
        import slune.videosequence as videosequence
        videosequence.Mission14Outro().start(
          tux   = videosequence.PLAYER,
          gnu   = videosequence.add_perso("gnu"  , 1),
          shark = videosequence.add_perso("shark", 4),
          )
        
      self.parent.remove(self)
      
      
    
  def advance_time(self, proportion):
    if self.parent:
      v = self >> self.receiver
      v.set_length(0.2 * proportion)
      self += v
      
      
class MedocWaiterMiniMap(widget.Widget):
  def __init__(self, level):
    self.level = level
    self.point = soya.Point()
    
    self.top = 10
    self.width  = 200
    self.height = 200
    self.left = soya.root_widget.width - self.width - 40
    
    flags = [e for e in level.children if isinstance(e, MedocWaiter)]
    min_x = min_y =  10e1000
    max_x = max_y = -10e1000
    for flag in flags:
      if flag.x < min_x: min_x = flag.x
      if flag.x > max_x: max_x = flag.x
      if flag.z < min_y: min_y = flag.z
      if flag.z > max_y: max_y = flag.z
    f_x = self.width  / (max_x - min_x)
    f_y = self.height / (max_y - min_y)
    if f_x < f_y:
      f = f_x
      left = self.left
    else:
      f = f_y
      left = self.left + ((max_y - min_y) - (max_x - min_x)) * f
    top = self.top
    
    self.f     = f
    self.cleft = left
    self.ctop  = top
    self.min_x = min_x
    self.min_y = min_y
    
    self.flags_pos = []
    self._calllist = soyaopengl.glGenList()
    soyaopengl.glNewList(self._calllist)
    soyaopengl.glColor4f(1.0, 1.0, 1.0, 1.0)
    soyaopengl.glBegin(soyaopengl.GL_QUADS)
    for flag in flags:
      x = left + f * (flag.x - min_x)
      y = top  + f * (flag.z - min_y)
      soyaopengl.glVertex2f(x - 4.0, y - 4.0)
      soyaopengl.glVertex2f(x - 4.0, y + 4.0)
      soyaopengl.glVertex2f(x + 4.0, y + 4.0)
      soyaopengl.glVertex2f(x + 4.0, y - 4.0)
    soyaopengl.glEnd()
    
    soyaopengl.glEndList()
    
  def resize(self, parent_left, parent_top, parent_width, parent_height):
    self.left = parent_width - self.width - 40
    self.screen_width  = parent_width
    self.screen_height = parent_height
    
  def render(self):
    soyaopengl.glCallList(self._calllist)
    
    for character in self.level.characters:
      if isinstance(character, slune_character.Competitor):
        x = self.cleft + self.f * (character.x - self.min_x)
        y = self.ctop  + self.f * (character.z - self.min_y)
        
        self.point.__init__(character, 0.0, 0.0, -25.0)
        self.point.convert_to(self.level)
        x2 = self.cleft + self.f * (self.point.x - self.min_x)
        y2 = self.ctop  + self.f * (self.point.z - self.min_y)
        
        soya.DEFAULT_MATERIAL.activate()
        soyaopengl.glBegin(soyaopengl.GL_LINES)
        soyaopengl.glVertex2f(x , y )
        soyaopengl.glVertex2f(x2, y2)
        soyaopengl.glEnd()
        
        material = soya.Material.get("head_" + character.perso_name + "_1")
        material.activate()
        soyaopengl.glEnable(soyaopengl.GL_BLEND)
        soyaopengl.glBegin(soyaopengl.GL_QUADS)
        soyaopengl.glTexCoord2f(0.0, 0.0); soyaopengl.glVertex2f(x - 20.0, y - 20.0)
        soyaopengl.glTexCoord2f(0.0, 1.0); soyaopengl.glVertex2f(x - 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 1.0); soyaopengl.glVertex2f(x + 20.0, y + 20.0)
        soyaopengl.glTexCoord2f(1.0, 0.0); soyaopengl.glVertex2f(x + 20.0, y - 20.0)
        soyaopengl.glEnd()
        soyaopengl.glDisable(soyaopengl.GL_BLEND)
        soya.DEFAULT_MATERIAL.activate()
  
        



# Slune
# Copyright (C) 2002-2003 Jean-Baptiste LAMY
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

import time, random, math

import soya
import py2play.level as py2p_level, py2play.action as action, py2play.character as character
import slune.character as slune_character
import slune.controler

CHROMOSOMS = {
  0: [0.9, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.1 ],
  1: [0.9, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 0.5],
  2: [0.9, 0.54152636319441949, 0.057433463934462713, 1.5669823329989161, 15.877046130951882, 0.18908021610553249, 0.066441050148349098, 3.0216818334258322, 1.0 ],
  }

class Praise(slune_character.NonPlayerCharacter):
  def __init__(self, parent = None, chromosom = None, path = (), escape_dist = 40.0, wait_dist = 10.0):
    slune_character.NonPlayerCharacter.__init__(self, parent)
    
    self.chromosom           = chromosom
    self.path                = path
    self.escape_dist         = escape_dist
    self.wait_dist           = wait_dist
    self.check_ground_vector = soya.Vector(self, 0.0, -1.5, -2.0)
    
  def finished(self): pass
  
  def escaped_from(self, hunter):
    import videosequence
    videosequence.HuntFailed().start(praise = self, hunter = hunter)
    
  def IAControler(character):
    c = character.chromosom = character.chromosom or CHROMOSOMS[character.level.difficulty]
    max_speed = c[8]
    jumping = 0
    
    mouse_x = mouse_y = 0.0
    
    while not character.path: yield slune_character.Action()
    
    while character.wait_dist:
      for hunter in character.level.characters:
        if isinstance(hunter, slune_character.Competitor):
          if character.distance_to(hunter) < character.wait_dist: character.wait_dist = 0.0
          yield slune_character.Action()
          
    while 1:
      if not character.path:
        while character.finished(): yield slune_character.Action(slune_character.ACTION_WAIT)
        while 1: yield slune_character.Action(slune_character.ACTION_WAIT)
        
      else:
        character.speed.y = 100.0 # Avoid null speed (null speed => jumping). The speed will be overrided later.
        
        next_flag = character.path.pop(0)
        
        if hasattr(next_flag, "next"): yield next_flag
        else:
          dist = 10.0
          while dist > 2.0:
            dist = character.distance_to(next_flag)
            for hunter in character.level.characters:
              if isinstance(hunter, slune_character.Competitor):
                if not(getattr(hunter, "failed", 0)) and (character.distance_to(hunter) > character.escape_dist):
                  hunter.failed = 1
                  character.escaped_from(hunter)
                  
            fx, fy, fz = character.transform_point(next_flag.x, next_flag.y, next_flag.z, next_flag.parent)
            
            if jumping:
              if character.on_ground: jumping = 0
              no_ground = 0 # It is normal that there is no ground !
            else:
              no_ground = not character.context.raypick_b(character, character.check_ground_vector, 4.0, 3)
              
            if ((dist < 12.0) and (c[5] * fy + c[6] * fz + c[7] * no_ground > 2.0)) or (character.speed.length() < 0.05):
              if character.random.random() < 0.7:
                action = slune_character.ACTION_JUMP
                jumping = 1
              else: yield slune.controler.Wall_climber(character)
            else:
              action = slune_character.ACTION_WAIT
              
            _mouse_x = min(1.0, -c[0] * math.sqrt(abs(fx / fz)))
            if fx > 0.0: _mouse_x = -_mouse_x
            mouse_x = 0.8 * mouse_x + 0.2 * _mouse_x
            mouse_y = (5.0 * mouse_y + min(max_speed, max(-max_speed, c[1] * abs(fx) + c[2] * fy + c[3] * fz + c[4] * no_ground))) / 6.0
            
            yield slune_character.Action(action, mouse_x, mouse_y)
            


class PathComputer:
  def __init__(self):
    self.pathes = {}
    
    
  def add_path(self, from_, to, path, reverse = 1):
    if self.pathes.has_key(from_):
      if self.pathes[from_].has_key(to):
        self.pathes[from_][to].append(path)
        
      else:
        self.pathes[from_][to] = [path]
        
    else:
      self.pathes[from_] = { to : [path] }
      
    if reverse:
      reverse_path = path[:]
      reverse_path.reverse()
      self.add_path(to, from_, reverse_path, 0)
      
  def random_path(self, start, end, approx_length, random = None):
    if not random: import random

    old_location = None
    location     = start
    
    path   = []
    length = 0
    while location != end:
      if length >= approx_length:
        if self.pathes[location].has_key(end):
          path.extend(random.choice(self.pathes[location][end]))
          #print "->", end
          
          break
        
      new_locations = self.pathes[location].keys()
      if (length < approx_length) and (len(new_locations) > 1):
        try: new_locations.remove(end)
        except: pass
        
      if len(new_locations) > 1:
        try: new_locations.remove(old_location)
        except: pass
        
      #print "choix parmi", new_locations
      
      old_location = location
      location     = random.choice(new_locations)
      path.extend(random.choice(self.pathes[old_location][location]))
      length += 1
      
      #print "->", location
      
    return path

  def path(self, *locations):
    path = []
    
    old_location = locations[0]
    for location in locations[1:]:
      path.extend(self.pathes[old_location][location][0])
      old_location = location
      
    return path
